"""Classes for docstring parsing and formatting."""

from __future__ import annotations

import collections
import contextlib
import inspect
import re
from functools import partial
from itertools import starmap
from typing import TYPE_CHECKING, Any

from sphinx.locale import _, __
from sphinx.util import logging
from sphinx.util.typing import get_type_hints, stringify_annotation

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import Literal

    from sphinx.application import Sphinx
    from sphinx.config import Config as SphinxConfig
    from sphinx.ext.autodoc._property_types import _AutodocObjType

logger = logging.getLogger(__name__)

_directive_regex = re.compile(r'\.\. \S+::')
_google_section_regex = re.compile(r'^(\s|\w)+:\s*$')
_google_typed_arg_regex = re.compile(r'(.+?)\(\s*(.*[^\s]+)\s*\)')
_numpy_section_regex = re.compile(r'^[=\-`:\'"~^_*+#<>]{2,}\s*$')
_single_colon_regex = re.compile(r'(?<!:):(?!:)')
_xref_or_code_regex = re.compile(
    r'((?::(?:[a-zA-Z0-9]+[\-_+:.])*[a-zA-Z0-9]+:`.+?`)|'
    r'(?:``.+?``)|'
    r'(?::meta .+:.*)|'
    r'(?:`.+?\s*(?<!\x00)<.*?>`))'
)
_xref_regex = re.compile(
    r'(?:(?::(?:[a-zA-Z0-9]+[\-_+:.])*[a-zA-Z0-9]+:)?`.+?`)',
)
_bullet_list_regex = re.compile(r'^(\*|\+|\-)(\s+\S|\s*$)')
_enumerated_list_regex = re.compile(
    r'^(?P<paren>\()?'
    r'(\d+|#|[ivxlcdm]+|[IVXLCDM]+|[a-zA-Z])'
    r'(?(paren)\)|\.)(\s+\S|\s*$)'
)
_token_regex = re.compile(
    r'(,\sor\s|\sor\s|\sof\s|:\s|\sto\s|,\sand\s|\sand\s|,\s'
    r'|[{]|[}]'
    r'|"(?:\\"|[^"])*"'
    r"|'(?:\\'|[^'])*')",
)
_default_regex = re.compile(
    r'^default[^_0-9A-Za-z].*$',
)
_SINGLETONS = frozenset({'None', 'True', 'False', 'Ellipsis', '...'})


class Deque(collections.deque[Any]):
    """A subclass of deque that mimics ``pockets.iterators.modify_iter``.

    The `.Deque.get` and `.Deque.next` methods are added.
    """

    sentinel = object()

    def get(self, n: int) -> Any:
        """Return the nth element of the stack, or ``self.sentinel`` if n is
        greater than the stack size.
        """
        return self[n] if n < len(self) else self.sentinel

    def next(self) -> Any:
        if self:
            return super().popleft()
        else:
            raise StopIteration


def _recombine_set_tokens(tokens: list[str]) -> list[str]:
    token_queue = collections.deque(tokens)
    keywords = ('optional', 'default')

    def takewhile_set(tokens: collections.deque[str]) -> Iterator[str]:
        open_braces = 0
        previous_token = None
        while True:
            try:
                token = tokens.popleft()
            except IndexError:
                break

            if token == ', ':
                previous_token = token
                continue

            if not token.strip():
                continue

            if token in keywords:
                tokens.appendleft(token)
                if previous_token is not None:
                    tokens.appendleft(previous_token)
                break

            if previous_token is not None:
                yield previous_token
                previous_token = None

            if token == '{':
                open_braces += 1
            elif token == '}':
                open_braces -= 1

            yield token

            if open_braces == 0:
                break

    def combine_set(tokens: collections.deque[str]) -> Iterator[str]:
        while True:
            try:
                token = tokens.popleft()
            except IndexError:
                break

            if token == '{':
                tokens.appendleft('{')
                yield ''.join(takewhile_set(tokens))
            else:
                yield token

    return list(combine_set(token_queue))


def _tokenize_type_spec(spec: str) -> list[str]:
    def postprocess(item: str) -> list[str]:
        if _default_regex.match(item):
            default = item[:7]
            # can't be separated by anything other than a single space
            # for now
            other = item[8:]

            return [default, ' ', other]
        else:
            return [item]

    tokens = [
        item
        for raw_token in _token_regex.split(spec)
        for item in postprocess(raw_token)
        if item
    ]
    return tokens


def _token_type(token: str, debug_location: str | None = None) -> str:
    def is_numeric(token: str) -> bool:
        try:
            # use complex to make sure every numeric value is detected as literal
            complex(token)
        except ValueError:
            return False
        else:
            return True

    if token.startswith(' ') or token.endswith(' '):
        type_ = 'delimiter'
    elif (
        is_numeric(token)
        or (token.startswith('{') and token.endswith('}'))
        or (token.startswith('"') and token.endswith('"'))
        or (token.startswith("'") and token.endswith("'"))
    ):
        type_ = 'literal'
    elif token.startswith('{'):
        logger.warning(
            __('invalid value set (missing closing brace): %s'),
            token,
            location=debug_location,
        )
        type_ = 'literal'
    elif token.endswith('}'):
        logger.warning(
            __('invalid value set (missing opening brace): %s'),
            token,
            location=debug_location,
        )
        type_ = 'literal'
    elif token.startswith(("'", '"')):
        logger.warning(
            __('malformed string literal (missing closing quote): %s'),
            token,
            location=debug_location,
        )
        type_ = 'literal'
    elif token.endswith(("'", '"')):
        logger.warning(
            __('malformed string literal (missing opening quote): %s'),
            token,
            location=debug_location,
        )
        type_ = 'literal'
    elif token in {'optional', 'default'}:
        # default is not a official keyword (yet) but supported by the
        # reference implementation (numpydoc) and widely used
        type_ = 'control'
    elif _xref_regex.match(token):
        type_ = 'reference'
    else:
        type_ = 'obj'

    return type_


def _convert_type_spec(
    _type: str,
    translations: dict[str, str] | None = None,
    debug_location: str | None = None,
) -> str:
    if translations is None:
        translations = {}

    tokens = _tokenize_type_spec(_type)
    combined_tokens = _recombine_set_tokens(tokens)
    types = [(token, _token_type(token, debug_location)) for token in combined_tokens]

    converters = {
        'literal': lambda x: f'``{x}``',
        'obj': lambda x: _convert_type_spec_obj(x, translations),
        'control': lambda x: f'*{x}*',
        'delimiter': lambda x: x,
        'reference': lambda x: x,
    }

    converted = ''.join(
        converters.get(type_)(token)  # type: ignore[misc]
        for token, type_ in types
    )

    return converted


def _convert_type_spec_obj(obj: str, translations: dict[str, str]) -> str:
    translation = translations.get(obj, obj)

    if _xref_regex.match(translation) is not None:
        return translation

    # use :py:obj: if obj is a standard singleton
    if translation in _SINGLETONS:
        if translation == '...':  # allow referencing the builtin ...
            return ':py:obj:`... <Ellipsis>`'
        return f':py:obj:`{translation}`'

    return f':py:class:`{translation}`'


class GoogleDocstring:
    """Convert Google style docstrings to reStructuredText.

    Parameters
    ----------
    docstring : :obj:`str` or :obj:`list` of :obj:`str`
        The docstring to parse, given either as a string or split into
        individual lines.
    config: :obj:`sphinx.ext.napoleon.Config` or :obj:`sphinx.config.Config`
        The configuration settings to use. If not given, defaults to the
        config object on `app`; or if `app` is not given defaults to the
        a new :class:`sphinx.ext.napoleon.Config` object.


    Other Parameters
    ----------------
    app : :class:`sphinx.application.Sphinx`, optional
        Application object representing the Sphinx process.
    what : :obj:`str`, optional
        A string specifying the type of the object to which the docstring
        belongs. Valid values: "module", "class", "exception", "function",
        "method", "attribute".
    name : :obj:`str`, optional
        The fully qualified name of the object.
    obj : module, class, exception, function, method, or attribute
        The object to which the docstring belongs.
    options : :class:`sphinx.ext.autodoc.Options`, optional
        The options given to the directive: an object with attributes
        inherited_members, undoc_members, show_inheritance and no_index that
        are True if the flag option of same name was given to the auto
        directive.


    Example
    -------
    >>> from sphinx.ext.napoleon import Config
    >>> config = Config(napoleon_use_param=True, napoleon_use_rtype=True)
    >>> docstring = '''One line summary.
    ...
    ... Extended description.
    ...
    ... Args:
    ...   arg1(int): Description of `arg1`
    ...   arg2(str): Description of `arg2`
    ... Returns:
    ...   str: Description of return value.
    ... '''
    >>> print(GoogleDocstring(docstring, config))
    One line summary.
    <BLANKLINE>
    Extended description.
    <BLANKLINE>
    :param arg1: Description of `arg1`
    :type arg1: int
    :param arg2: Description of `arg2`
    :type arg2: str
    <BLANKLINE>
    :returns: Description of return value.
    :rtype: str
    <BLANKLINE>

    """

    _name_rgx = re.compile(
        r'^\s*((?::(?P<role>\S+):)?`(?P<name>~?[a-zA-Z0-9_.-]+)`|'
        r' (?P<name2>~?[a-zA-Z0-9_.-]+))\s*',
        re.VERBOSE,
    )

    def __init__(
        self,
        docstring: str | list[str],
        config: SphinxConfig | None = None,
        app: Sphinx | None = None,
        what: _AutodocObjType | Literal['object'] = 'object',
        name: str = '',
        obj: Any | None = None,
        options: Any | None = None,
    ) -> None:
        self._app = app
        if config:
            self._config = config
        elif app:
            self._config = app.config
        else:
            from sphinx.ext.napoleon import Config

            self._config = Config()  # type: ignore[assignment]

        if what == 'object':
            if inspect.isclass(obj):
                what = 'class'
            elif inspect.ismodule(obj):
                what = 'module'
            elif callable(obj):
                what = 'function'
            else:
                what = 'object'

        self._what: _AutodocObjType | Literal['object'] = what
        self._name = name
        self._obj = obj
        if options:
            try:
                self._no_index = options.no_index or options.noindex
            except (AttributeError, TypeError):
                self._no_index = False
        else:
            self._no_index = False
        if isinstance(docstring, str):
            lines = docstring.splitlines()
        else:
            lines = docstring
        self._lines = Deque(map(str.rstrip, lines))
        self._parsed_lines: list[str] = []
        self._is_in_section = False
        self._section_indent = 0
        if not hasattr(self, '_directive_sections'):
            self._directive_sections: list[str] = []
        if not hasattr(self, '_sections'):
            self._sections: dict[str, Callable[..., list[str]]] = {
                'args': self._parse_parameters_section,
                'arguments': self._parse_parameters_section,
                'attention': partial(self._parse_admonition, 'attention'),
                'attributes': self._parse_attributes_section,
                'caution': partial(self._parse_admonition, 'caution'),
                'danger': partial(self._parse_admonition, 'danger'),
                'error': partial(self._parse_admonition, 'error'),
                'example': self._parse_examples_section,
                'examples': self._parse_examples_section,
                'hint': partial(self._parse_admonition, 'hint'),
                'important': partial(self._parse_admonition, 'important'),
                'keyword args': self._parse_keyword_arguments_section,
                'keyword arguments': self._parse_keyword_arguments_section,
                'methods': self._parse_methods_section,
                'note': partial(self._parse_admonition, 'note'),
                'notes': self._parse_notes_section,
                'other parameters': self._parse_other_parameters_section,
                'parameters': self._parse_parameters_section,
                'receive': self._parse_receives_section,
                'receives': self._parse_receives_section,
                'return': self._parse_returns_section,
                'returns': self._parse_returns_section,
                'raise': self._parse_raises_section,
                'raises': self._parse_raises_section,
                'references': self._parse_references_section,
                'see also': self._parse_see_also_section,
                'tip': partial(self._parse_admonition, 'tip'),
                'todo': partial(self._parse_admonition, 'todo'),
                'warning': partial(self._parse_admonition, 'warning'),
                'warnings': partial(self._parse_admonition, 'warning'),
                'warn': self._parse_warns_section,
                'warns': self._parse_warns_section,
                'yield': self._parse_yields_section,
                'yields': self._parse_yields_section,
            }

        self._load_custom_sections()

        self._parse()

    def __str__(self) -> str:
        """Return the parsed docstring in reStructuredText format.

        Returns
        -------
        unicode
            Unicode version of the docstring.

        """
        return '\n'.join(self.lines())

    def _get_location(self) -> str | None:
        try:
            filepath = inspect.getfile(self._obj) if self._obj is not None else None
        except TypeError:
            filepath = None
        name = self._name

        if filepath is None and name is None:
            return None
        elif filepath is None:
            filepath = ''

        return f'{filepath}:docstring of {name}'

    def lines(self) -> list[str]:
        """Return the parsed lines of the docstring in reStructuredText format.

        Returns
        -------
        list(str)
            The lines of the docstring in a list.

        """
        return self._parsed_lines

    def _consume_indented_block(self, indent: int = 1) -> list[str]:
        lines = []
        line = self._lines.get(0)
        while not self._is_section_break() and (
            not line or self._is_indented(line, indent)
        ):
            lines.append(self._lines.next())
            line = self._lines.get(0)
        return lines

    def _consume_contiguous(self) -> list[str]:
        lines = []
        while self._lines and self._lines.get(0) and not self._is_section_header():
            lines.append(self._lines.next())
        return lines

    def _consume_empty(self) -> list[str]:
        lines = []
        line = self._lines.get(0)
        while self._lines and not line:
            lines.append(self._lines.next())
            line = self._lines.get(0)
        return lines

    def _consume_field(
        self,
        parse_type: bool = True,
        prefer_type: bool = False,
    ) -> tuple[str, str, list[str]]:
        line = self._lines.next()

        before, _colon, after = self._partition_field_on_colon(line)
        _name, _type, _desc = before, '', after

        if parse_type:
            match = _google_typed_arg_regex.match(before)
            if match:
                _name = match.group(1).strip()
                _type = match.group(2)

        _name = self._escape_args_and_kwargs(_name)

        if prefer_type and not _type:
            _type, _name = _name, _type

        if _type and self._config.napoleon_preprocess_types:
            _type = _convert_type_spec(
                _type,
                translations=self._config.napoleon_type_aliases or {},
                debug_location=self._get_location(),
            )

        indent = self._get_indent(line) + 1
        _descs = [_desc, *self._dedent(self._consume_indented_block(indent))]
        _descs = self.__class__(_descs, self._config).lines()
        return _name, _type, _descs

    def _consume_fields(
        self, parse_type: bool = True, prefer_type: bool = False, multiple: bool = False
    ) -> list[tuple[str, str, list[str]]]:
        self._consume_empty()
        fields: list[tuple[str, str, list[str]]] = []
        while not self._is_section_break():
            _name, _type, _desc = self._consume_field(parse_type, prefer_type)
            if multiple and _name:
                fields.extend((name.strip(), _type, _desc) for name in _name.split(','))
            elif _name or _type or _desc:
                fields.append((_name, _type, _desc))
        return fields

    def _consume_inline_attribute(self) -> tuple[str, list[str]]:
        line = self._lines.next()
        _type, colon, _desc = self._partition_field_on_colon(line)
        if not colon or not _desc:
            _type, _desc = _desc, _type
            _desc += colon
        _descs = [_desc, *self._dedent(self._consume_to_end())]
        _descs = self.__class__(_descs, self._config).lines()
        return _type, _descs

    def _consume_returns_section(
        self, preprocess_types: bool = False
    ) -> list[tuple[str, str, list[str]]]:
        lines = self._dedent(self._consume_to_next_section())
        if lines:
            before, colon, after = self._partition_field_on_colon(lines[0])
            _name, _type, _desc = '', '', lines

            if colon:
                if after:
                    _desc = [after, *lines[1:]]
                else:
                    _desc = lines[1:]

                _type = before

            if _type and preprocess_types and self._config.napoleon_preprocess_types:
                _type = _convert_type_spec(
                    _type,
                    translations=self._config.napoleon_type_aliases or {},
                    debug_location=self._get_location(),
                )

            _desc = self.__class__(_desc, self._config).lines()
            return [(_name, _type, _desc)]
        else:
            return []

    def _consume_usage_section(self) -> list[str]:
        lines = self._dedent(self._consume_to_next_section())
        return lines

    def _consume_section_header(self) -> str:
        section = self._lines.next()
        stripped_section = section.strip(':')
        if stripped_section.lower() in self._sections:
            section = stripped_section
        return section

    def _consume_to_end(self) -> list[str]:
        lines = []
        while self._lines:
            lines.append(self._lines.next())
        return lines

    def _consume_to_next_section(self) -> list[str]:
        self._consume_empty()
        lines = []
        while not self._is_section_break():
            lines.append(self._lines.next())
        return lines + self._consume_empty()

    def _dedent(self, lines: list[str], full: bool = False) -> list[str]:
        if full:
            return [line.lstrip() for line in lines]
        else:
            min_indent = self._get_min_indent(lines)
            return [line[min_indent:] for line in lines]

    def _escape_args_and_kwargs(self, name: str) -> str:
        if name.endswith('_') and getattr(
            self._config, 'strip_signature_backslash', False
        ):
            name = name[:-1] + r'\_'

        if name[:2] == '**':
            return r'\*\*' + name[2:]
        elif name[:1] == '*':
            return r'\*' + name[1:]
        else:
            return name

    def _fix_field_desc(self, desc: list[str]) -> list[str]:
        if self._is_list(desc):
            desc = ['', *desc]
        elif desc[0].endswith('::'):
            desc_block = desc[1:]
            indent = self._get_indent(desc[0])
            block_indent = self._get_initial_indent(desc_block)
            if block_indent > indent:
                desc = ['', *desc]
            else:
                desc = ['', desc[0], *self._indent(desc_block, 4)]
        return desc

    def _format_admonition(self, admonition: str, lines: list[str]) -> list[str]:
        lines = self._strip_empty(lines)
        if len(lines) == 1:
            return [f'.. {admonition}:: {lines[0].strip()}', '']
        elif lines:
            lines = self._indent(self._dedent(lines), 3)
            return [f'.. {admonition}::', '', *lines, '']
        else:
            return [f'.. {admonition}::', '']

    def _format_block(
        self,
        prefix: str,
        lines: list[str],
        padding: str | None = None,
    ) -> list[str]:
        if lines:
            if padding is None:
                padding = ' ' * len(prefix)
            result_lines = []
            for i, line in enumerate(lines):
                if i == 0:
                    result_lines.append((prefix + line).rstrip())
                elif line:
                    result_lines.append(padding + line)
                else:
                    result_lines.append('')
            return result_lines
        else:
            return [prefix]

    def _format_docutils_params(
        self,
        fields: list[tuple[str, str, list[str]]],
        field_role: str = 'param',
        type_role: str = 'type',
    ) -> list[str]:
        lines = []
        for _name, _type, _desc in fields:
            _desc = self._strip_empty(_desc)
            if any(_desc):
                _desc = self._fix_field_desc(_desc)
                field = f':{field_role} {_name}: '
                lines.extend(self._format_block(field, _desc))
            else:
                lines.append(f':{field_role} {_name}:')

            if _type:
                lines.append(f':{type_role} {_name}: {_type}')
        return [*lines, '']

    def _format_field(self, _name: str, _type: str, _desc: list[str]) -> list[str]:
        _desc = self._strip_empty(_desc)
        has_desc = any(_desc)
        separator = ' -- ' if has_desc else ''
        if _name:
            if _type:
                if '`' in _type:
                    field = f'**{_name}** ({_type}){separator}'
                else:
                    field = f'**{_name}** (*{_type}*){separator}'
            else:
                field = f'**{_name}**{separator}'
        elif _type:
            if '`' in _type:
                field = f'{_type}{separator}'
            else:
                field = f'*{_type}*{separator}'
        else:
            field = ''

        if has_desc:
            _desc = self._fix_field_desc(_desc)
            if _desc[0]:
                return [field + _desc[0], *_desc[1:]]
            else:
                return [field, *_desc]
        else:
            return [field]

    def _format_fields(
        self,
        field_type: str,
        fields: list[tuple[str, str, list[str]]],
    ) -> list[str]:
        field_type = f':{field_type.strip()}:'
        padding = ' ' * len(field_type)
        multi = len(fields) > 1
        lines: list[str] = []
        for _name, _type, _desc in fields:
            field = self._format_field(_name, _type, _desc)
            if multi:
                if lines:
                    lines.extend(self._format_block(padding + ' * ', field))
                else:
                    lines.extend(self._format_block(field_type + ' * ', field))
            else:
                lines.extend(self._format_block(field_type + ' ', field))
        if lines and lines[-1]:
            lines.append('')
        return lines

    def _get_current_indent(self, peek_ahead: int = 0) -> int:
        line = self._lines.get(peek_ahead)
        while line is not self._lines.sentinel:
            if line:
                return self._get_indent(line)
            peek_ahead += 1
            line = self._lines.get(peek_ahead)
        return 0

    def _get_indent(self, line: str) -> int:
        for i, s in enumerate(line):
            if not s.isspace():
                return i
        return len(line)

    def _get_initial_indent(self, lines: list[str]) -> int:
        for line in lines:
            if line:
                return self._get_indent(line)
        return 0

    def _get_min_indent(self, lines: list[str]) -> int:
        min_indent = None
        for line in lines:
            if line:
                indent = self._get_indent(line)
                if min_indent is None or indent < min_indent:
                    min_indent = indent
        return min_indent or 0

    def _indent(self, lines: list[str], n: int = 4) -> list[str]:
        return [(' ' * n) + line for line in lines]

    def _is_indented(self, line: str, indent: int = 1) -> bool:
        for i, s in enumerate(line):
            if i >= indent:
                return True
            elif not s.isspace():
                return False
        return False

    def _is_list(self, lines: list[str]) -> bool:
        if not lines:
            return False
        if _bullet_list_regex.match(lines[0]):
            return True
        if _enumerated_list_regex.match(lines[0]):
            return True
        if len(lines) < 2 or lines[0].endswith('::'):
            return False
        indent = self._get_indent(lines[0])
        next_indent = indent
        for line in lines[1:]:
            if line:
                next_indent = self._get_indent(line)
                break
        return next_indent > indent

    def _is_section_header(self) -> bool:
        section = self._lines.get(0).lower()
        match = _google_section_regex.match(section)
        if match and section.strip(':') in self._sections:
            header_indent = self._get_indent(section)
            section_indent = self._get_current_indent(peek_ahead=1)
            return section_indent > header_indent
        elif self._directive_sections:
            if _directive_regex.match(section):
                for directive_section in self._directive_sections:
                    if section.startswith(directive_section):
                        return True
        return False

    def _is_section_break(self) -> bool:
        line = self._lines.get(0)
        return (
            not self._lines
            or self._is_section_header()
            or (
                self._is_in_section
                and line
                and not self._is_indented(line, self._section_indent)
            )
        )

    def _load_custom_sections(self) -> None:
        if self._config.napoleon_custom_sections is not None:
            for entry in self._config.napoleon_custom_sections:
                if isinstance(entry, str):
                    # if entry is just a label, add to sections list,
                    # using generic section logic.
                    self._sections[entry.lower()] = self._parse_custom_generic_section
                else:
                    # otherwise, assume entry is container;
                    if entry[1] == 'params_style':
                        self._sections[entry[0].lower()] = (
                            self._parse_custom_params_style_section
                        )
                    elif entry[1] == 'returns_style':
                        self._sections[entry[0].lower()] = (
                            self._parse_custom_returns_style_section
                        )
                    else:
                        # [0] is new section, [1] is the section to alias.
                        # in the case of key mismatch, just handle as generic section.
                        self._sections[entry[0].lower()] = self._sections.get(
                            entry[1].lower(), self._parse_custom_generic_section
                        )

    def _parse(self) -> None:
        self._parsed_lines = self._consume_empty()

        if self._name and self._what in {'attribute', 'data', 'property'}:
            res: list[str] = []
            with contextlib.suppress(StopIteration):
                res = self._parse_attribute_docstring()

            self._parsed_lines.extend(res)
            return

        while self._lines:
            if self._is_section_header():
                try:
                    section = self._consume_section_header()
                    self._is_in_section = True
                    self._section_indent = self._get_current_indent()
                    if _directive_regex.match(section):
                        lines = [section, *self._consume_to_next_section()]
                    else:
                        lines = self._sections[section.lower()](section)
                finally:
                    self._is_in_section = False
                    self._section_indent = 0
            else:
                if not self._parsed_lines:
                    lines = self._consume_contiguous() + self._consume_empty()
                else:
                    lines = self._consume_to_next_section()
            self._parsed_lines.extend(lines)

    def _parse_admonition(self, admonition: str, section: str) -> list[str]:
        # type (str, str) -> List[str]
        lines = self._consume_to_next_section()
        return self._format_admonition(admonition, lines)

    def _parse_attribute_docstring(self) -> list[str]:
        _type, _desc = self._consume_inline_attribute()
        lines = self._format_field('', '', _desc)
        if _type:
            lines.extend(['', f':type: {_type}'])
        return lines

    def _parse_attributes_section(self, section: str) -> list[str]:
        lines = []
        for _name, _type, _desc in self._consume_fields():
            if not _type:
                _type = self._lookup_annotation(_name)
            if self._config.napoleon_use_ivar:
                field = f':ivar {_name}: '
                lines.extend(self._format_block(field, _desc))
                if _type:
                    lines.append(f':vartype {_name}: {_type}')
            else:
                lines.append('.. attribute:: ' + _name)
                if self._no_index:
                    lines.append('   :no-index:')
                lines.append('')

                fields = self._format_field('', '', _desc)
                lines.extend(self._indent(fields, 3))
                if _type:
                    lines.append('')
                    lines.extend(self._indent([f':type: {_type}'], 3))
                lines.append('')
        if self._config.napoleon_use_ivar:
            lines.append('')
        return lines

    def _parse_examples_section(self, section: str) -> list[str]:
        labels = {
            'example': _('Example'),
            'examples': _('Examples'),
        }
        use_admonition = self._config.napoleon_use_admonition_for_examples
        label = labels.get(section.lower(), section)
        return self._parse_generic_section(label, use_admonition)

    def _parse_custom_generic_section(self, section: str) -> list[str]:
        # for now, no admonition for simple custom sections
        return self._parse_generic_section(section, False)

    def _parse_custom_params_style_section(self, section: str) -> list[str]:
        return self._format_fields(section, self._consume_fields())

    def _parse_custom_returns_style_section(self, section: str) -> list[str]:
        fields = self._consume_returns_section(preprocess_types=True)
        return self._format_fields(section, fields)

    def _parse_usage_section(self, section: str) -> list[str]:
        header = ['.. rubric:: Usage:', '']
        block = ['.. code-block:: python', '']
        lines = self._consume_usage_section()
        lines = self._indent(lines, 3)
        return header + block + lines + ['']

    def _parse_generic_section(self, section: str, use_admonition: bool) -> list[str]:
        lines = self._strip_empty(self._consume_to_next_section())
        lines = self._dedent(lines)
        if use_admonition:
            header = f'.. admonition:: {section}'
            lines = self._indent(lines, 3)
        else:
            header = f'.. rubric:: {section}'
        if lines:
            return [header, '', *lines, '']
        else:
            return [header, '']

    def _parse_keyword_arguments_section(self, section: str) -> list[str]:
        fields = self._consume_fields()
        if self._config.napoleon_use_keyword:
            return self._format_docutils_params(
                fields, field_role='keyword', type_role='kwtype'
            )
        else:
            return self._format_fields(_('Keyword Arguments'), fields)

    def _parse_methods_section(self, section: str) -> list[str]:
        lines: list[str] = []
        for _name, _type, _desc in self._consume_fields(parse_type=False):
            lines.append(f'.. method:: {_name}')
            if self._no_index:
                lines.append('   :no-index:')
            if _desc:
                lines.extend(['', *self._indent(_desc, 3)])
            lines.append('')
        return lines

    def _parse_notes_section(self, section: str) -> list[str]:
        use_admonition = self._config.napoleon_use_admonition_for_notes
        return self._parse_generic_section(_('Notes'), use_admonition)

    def _parse_other_parameters_section(self, section: str) -> list[str]:
        if self._config.napoleon_use_param:
            # Allow to declare multiple parameters at once (ex: x, y: int)
            fields = self._consume_fields(multiple=True)
            return self._format_docutils_params(fields)
        else:
            fields = self._consume_fields()
            return self._format_fields(_('Other Parameters'), fields)

    def _parse_parameters_section(self, section: str) -> list[str]:
        if self._config.napoleon_use_param:
            # Allow to declare multiple parameters at once (ex: x, y: int)
            fields = self._consume_fields(multiple=True)
            return self._format_docutils_params(fields)
        else:
            fields = self._consume_fields()
            return self._format_fields(_('Parameters'), fields)

    def _parse_raises_section(self, section: str) -> list[str]:
        fields = self._consume_fields(parse_type=False, prefer_type=True)
        lines: list[str] = []
        for _name, _type, _desc in fields:
            m = self._name_rgx.match(_type)
            if m and m.group('name'):
                _type = m.group('name')
            elif _xref_regex.match(_type):
                pos = _type.find('`')
                _type = _type[pos + 1 : -1]
            _type = ' ' + _type if _type else ''
            _desc = self._strip_empty(_desc)
            _descs = ' ' + '\n    '.join(_desc) if any(_desc) else ''
            lines.append(f':raises{_type}:{_descs}')
        if lines:
            lines.append('')
        return lines

    def _parse_receives_section(self, section: str) -> list[str]:
        if self._config.napoleon_use_param:
            # Allow to declare multiple parameters at once (ex: x, y: int)
            fields = self._consume_fields(multiple=True)
            return self._format_docutils_params(fields)
        else:
            fields = self._consume_fields()
            return self._format_fields(_('Receives'), fields)

    def _parse_references_section(self, section: str) -> list[str]:
        use_admonition = self._config.napoleon_use_admonition_for_references
        return self._parse_generic_section(_('References'), use_admonition)

    def _parse_returns_section(self, section: str) -> list[str]:
        fields = self._consume_returns_section()
        multi = len(fields) > 1
        use_rtype = False if multi else self._config.napoleon_use_rtype
        lines: list[str] = []

        for _name, _type, _desc in fields:
            if use_rtype:
                field = self._format_field(_name, '', _desc)
            else:
                field = self._format_field(_name, _type, _desc)

            if multi:
                if lines:
                    lines.extend(self._format_block('          * ', field))
                else:
                    lines.extend(self._format_block(':returns: * ', field))
            else:
                if any(field):  # only add :returns: if there's something to say
                    lines.extend(self._format_block(':returns: ', field))
                if _type and use_rtype:
                    lines.extend([f':rtype: {_type}', ''])
        if lines and lines[-1]:
            lines.append('')
        return lines

    def _parse_see_also_section(self, section: str) -> list[str]:
        return self._parse_admonition('seealso', section)

    def _parse_warns_section(self, section: str) -> list[str]:
        return self._format_fields(_('Warns'), self._consume_fields())

    def _parse_yields_section(self, section: str) -> list[str]:
        fields = self._consume_returns_section(preprocess_types=True)
        return self._format_fields(_('Yields'), fields)

    def _partition_field_on_colon(self, line: str) -> tuple[str, str, str]:
        before_colon = []
        after_colon = []
        colon = ''
        found_colon = False
        for i, source in enumerate(_xref_or_code_regex.split(line)):
            if found_colon:
                after_colon.append(source)
            else:
                m = _single_colon_regex.search(source)
                if (i % 2) == 0 and m:
                    found_colon = True
                    colon = source[m.start() : m.end()]
                    before_colon.append(source[: m.start()])
                    after_colon.append(source[m.end() :])
                else:
                    before_colon.append(source)

        return ''.join(before_colon).strip(), colon, ''.join(after_colon).strip()

    def _strip_empty(self, lines: list[str]) -> list[str]:
        if lines:
            start = -1
            for i, line in enumerate(lines):
                if line:
                    start = i
                    break
            if start == -1:
                lines = []
            end = -1
            for i in reversed(range(len(lines))):
                line = lines[i]
                if line:
                    end = i
                    break
            if start > 0 or end + 1 < len(lines):
                lines = lines[start : end + 1]
        return lines

    def _lookup_annotation(self, _name: str) -> str:
        if self._config.napoleon_attr_annotations:
            if self._what in {'module', 'class', 'exception'} and self._obj:
                # cache the class annotations
                if not hasattr(self, '_annotations'):
                    localns = getattr(self._config, 'autodoc_type_aliases', {})
                    localns.update(
                        getattr(
                            self._config,
                            'napoleon_type_aliases',
                            {},
                        )
                        or {}
                    )
                    self._annotations = get_type_hints(self._obj, None, localns)
                if _name in self._annotations:
                    short_literals = getattr(
                        self._config, 'python_display_short_literal_types', False
                    )
                    return stringify_annotation(
                        self._annotations[_name],
                        mode='fully-qualified-except-typing',
                        short_literals=short_literals,
                    )
        # No annotation found
        return ''


class NumpyDocstring(GoogleDocstring):
    """Convert NumPy style docstrings to reStructuredText.

    Parameters
    ----------
    docstring : :obj:`str` or :obj:`list` of :obj:`str`
        The docstring to parse, given either as a string or split into
        individual lines.
    config: :obj:`sphinx.ext.napoleon.Config` or :obj:`sphinx.config.Config`
        The configuration settings to use. If not given, defaults to the
        config object on `app`; or if `app` is not given defaults to the
        a new :class:`sphinx.ext.napoleon.Config` object.


    Other Parameters
    ----------------
    app : :class:`sphinx.application.Sphinx`, optional
        Application object representing the Sphinx process.
    what : :obj:`str`, optional
        A string specifying the type of the object to which the docstring
        belongs. Valid values: "module", "class", "exception", "function",
        "method", "attribute".
    name : :obj:`str`, optional
        The fully qualified name of the object.
    obj : module, class, exception, function, method, or attribute
        The object to which the docstring belongs.
    options : :class:`sphinx.ext.autodoc.Options`, optional
        The options given to the directive: an object with attributes
        inherited_members, undoc_members, show_inheritance and no_index that
        are True if the flag option of same name was given to the auto
        directive.


    Example
    -------
    >>> from sphinx.ext.napoleon import Config
    >>> config = Config(napoleon_use_param=True, napoleon_use_rtype=True)
    >>> docstring = '''One line summary.
    ...
    ... Extended description.
    ...
    ... Parameters
    ... ----------
    ... arg1 : int
    ...     Description of `arg1`
    ... arg2 : str
    ...     Description of `arg2`
    ... Returns
    ... -------
    ... str
    ...     Description of return value.
    ... '''
    >>> print(NumpyDocstring(docstring, config))
    One line summary.
    <BLANKLINE>
    Extended description.
    <BLANKLINE>
    :param arg1: Description of `arg1`
    :type arg1: int
    :param arg2: Description of `arg2`
    :type arg2: str
    <BLANKLINE>
    :returns: Description of return value.
    :rtype: str
    <BLANKLINE>

    Methods
    -------
    __str__()
        Return the parsed docstring in reStructuredText format.

        Returns
        -------
        str
            UTF-8 encoded version of the docstring.

    __unicode__()
        Return the parsed docstring in reStructuredText format.

        Returns
        -------
        unicode
            Unicode version of the docstring.

    lines()
        Return the parsed lines of the docstring in reStructuredText format.

        Returns
        -------
        list(str)
            The lines of the docstring in a list.

    """

    def __init__(
        self,
        docstring: str | list[str],
        config: SphinxConfig | None = None,
        app: Sphinx | None = None,
        what: _AutodocObjType | Literal['object'] = 'object',
        name: str = '',
        obj: Any | None = None,
        options: Any | None = None,
    ) -> None:
        self._directive_sections = ['.. index::']
        super().__init__(docstring, config, app, what, name, obj, options)

    def _escape_args_and_kwargs(self, name: str) -> str:
        func = super()._escape_args_and_kwargs

        if ', ' in name:
            return ', '.join(map(func, name.split(', ')))
        else:
            return func(name)

    def _consume_field(
        self, parse_type: bool = True, prefer_type: bool = False
    ) -> tuple[str, str, list[str]]:
        line = self._lines.next()
        if parse_type:
            _name, _, _type = self._partition_field_on_colon(line)
        else:
            _name, _type = line, ''
        _name, _type = _name.strip(), _type.strip()
        _name = self._escape_args_and_kwargs(_name)

        if parse_type and not _type:
            _type = self._lookup_annotation(_name)

        if prefer_type and not _type:
            _type, _name = _name, _type

        if self._config.napoleon_preprocess_types:
            _type = _convert_type_spec(
                _type,
                translations=self._config.napoleon_type_aliases or {},
                debug_location=self._get_location(),
            )

        indent = self._get_indent(line) + 1
        _desc = self._dedent(self._consume_indented_block(indent))
        _desc = self.__class__(_desc, self._config).lines()
        return _name, _type, _desc

    def _consume_returns_section(
        self, preprocess_types: bool = False
    ) -> list[tuple[str, str, list[str]]]:
        return self._consume_fields(prefer_type=True)

    def _consume_section_header(self) -> str:
        section = self._lines.next()
        if not _directive_regex.match(section):
            # Consume the header underline
            self._lines.next()
        return section

    def _is_section_break(self) -> bool:
        line1, line2 = self._lines.get(0), self._lines.get(1)
        return (
            not self._lines
            or self._is_section_header()
            or (not line1 and not line2)
            or (
                self._is_in_section
                and line1
                and not self._is_indented(line1, self._section_indent)
            )
        )

    def _is_section_header(self) -> bool:
        section, underline = self._lines.get(0), self._lines.get(1)
        section = section.lower()
        if section in self._sections and isinstance(underline, str):
            return bool(_numpy_section_regex.match(underline))
        elif self._directive_sections:
            if _directive_regex.match(section):
                for directive_section in self._directive_sections:
                    if section.startswith(directive_section):
                        return True
        return False

    def _parse_see_also_section(self, section: str) -> list[str]:
        lines = self._consume_to_next_section()
        try:
            return self._parse_numpydoc_see_also_section(lines)
        except ValueError:
            return self._format_admonition('seealso', lines)

    def _parse_numpydoc_see_also_section(self, content: list[str]) -> list[str]:
        """See Also
        --------
        func_name : Descriptive text
            continued text
        another_func_name : Descriptive text
        func_name1, func_name2, :meth:`func_name`, func_name3

        Licence
        -------

        Derived from the NumpyDoc implementation of ``_parse_see_also``,
        which was under the following licence:

            Copyright (C) 2008 Stefan van der Walt <stefan@mentat.za.net>,
                               Pauli Virtanen <pav@iki.fi>

            Redistribution and use in source and binary forms, with or without
            modification, are permitted provided that the following conditions are
            met:

             1. Redistributions of source code must retain the above copyright
                notice, this list of conditions and the following disclaimer.
             2. Redistributions in binary form must reproduce the above copyright
                notice, this list of conditions and the following disclaimer in
                the documentation and/or other materials provided with the
                distribution.

            THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
            IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
            WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
            DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
            INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
            (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
            SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
            HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
            STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
            IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
            POSSIBILITY OF SUCH DAMAGE.
        """
        items: list[tuple[str, list[str], str | None]] = []

        def parse_item_name(text: str) -> tuple[str, str | None]:
            """Match ':role:`name`' or 'name'"""
            m = self._name_rgx.match(text)
            if m:
                g = m.groups()
                if g[1] is None:
                    return g[3], None
                else:
                    return g[2], g[1]
            msg = f'{text} is not a item name'
            raise ValueError(msg)

        def push_item(name: str | None, rest: list[str]) -> None:
            if not name:
                return
            name, role = parse_item_name(name)
            items.append((name, rest.copy(), role))
            rest.clear()

        def translate(
            func: str,
            description: list[str],
            role: str | None,
        ) -> tuple[str, list[str], str | None]:
            translations = self._config.napoleon_type_aliases
            if role is not None or not translations:
                return func, description, role

            translated = translations.get(func, func)
            match = self._name_rgx.match(translated)
            if not match:
                return translated, description, role

            groups = match.groupdict()
            role = groups['role']
            new_func = groups['name'] or groups['name2']

            return new_func, description, role

        current_func = None
        rest: list[str] = []

        for line in content:
            if not line.strip():
                continue

            m = self._name_rgx.match(line)
            if m and line[m.end() :].strip().startswith(':'):
                push_item(current_func, rest)
                current_func, line = line[: m.end()], line[m.end() :]
                rest = [line.partition(':')[-1].strip()]
                if not rest[0]:
                    rest = []
            elif not line.startswith(' '):
                push_item(current_func, rest)
                current_func = None
                if ',' in line:
                    for func in line.split(','):
                        if func.strip():
                            push_item(func, [])
                elif line.strip():
                    current_func = line
            elif current_func is not None:
                rest.append(line.strip())
        push_item(current_func, rest)

        if not items:
            return []

        # apply type aliases
        items = list(starmap(translate, items))

        lines: list[str] = []
        last_had_desc = True
        for name, desc, role in items:
            if role:
                link = f':{role}:`{name}`'
            else:
                link = f':py:obj:`{name}`'
            if desc or last_had_desc:
                lines += ['']
                lines += [link]
            else:
                lines[-1] += f', {link}'
            if desc:
                lines += self._indent([' '.join(desc)])
                last_had_desc = True
            else:
                last_had_desc = False
        lines += ['']

        return self._format_admonition('seealso', lines)



from __future__ import annotations

import operator
import re
from enum import Enum
from typing import TYPE_CHECKING, Literal, NewType, TypeVar

from sphinx.errors import PycodeError
from sphinx.events import EventManager
from sphinx.ext.autodoc._directive_options import _AutoDocumenterOptions
from sphinx.ext.autodoc._dynamic._loader import _load_object_by_name
from sphinx.ext.autodoc._dynamic._mock import ismock, undecorate
from sphinx.ext.autodoc._property_types import _ClassDefProperties, _ModuleProperties
from sphinx.ext.autodoc._sentinels import ALL, INSTANCE_ATTR, SLOTS_ATTR
from sphinx.ext.autodoc._shared import LOGGER
from sphinx.locale import __
from sphinx.pycode import ModuleAnalyzer
from sphinx.util import inspect
from sphinx.util.docstrings import separate_metadata
from sphinx.util.inspect import (
    getannotations,
    getdoc,
    getmro,
    getslots,
    isclass,
    isenumclass,
    safe_getattr,
    unwrap_all,
)
from sphinx.util.typing import AnyTypeAliasType

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, MutableSet, Sequence, Set
    from typing import Any, Literal

    from sphinx.environment import _CurrentDocument
    from sphinx.events import EventManager
    from sphinx.ext.autodoc._directive_options import _AutoDocumenterOptions
    from sphinx.ext.autodoc._property_types import _AutodocObjType, _ItemProperties
    from sphinx.ext.autodoc._sentinels import (
        ALL_T,
        EMPTY_T,
        INSTANCE_ATTR_T,
        SLOTS_ATTR_T,
    )
    from sphinx.ext.autodoc._shared import _AttrGetter, _AutodocConfig

special_member_re = re.compile(r'^__\S+__$')


class ObjectMember:
    """A member of object.

    This is used for the result of `_get_members_to_document()` to
    represent each member of the object.
    """

    __slots__ = '__name__', 'object', 'docstring', 'class_'

    __name__: str
    object: Any
    docstring: Sequence[str] | None
    class_: Any
    skipped: bool

    def __init__(
        self,
        name: str,
        obj: INSTANCE_ATTR_T | SLOTS_ATTR_T | Any,
        *,
        docstring: Sequence[str] | None = None,
        class_: Any = None,
    ) -> None:
        self.__name__ = name
        self.object = obj
        self.docstring = docstring
        self.class_ = class_

    def __repr__(self) -> str:
        return (
            f'ObjectMember('
            f'name={self.__name__!r}, '
            f'obj={self.object!r}, '
            f'docstring={self.docstring!r}, '
            f'class_={self.class_!r}'
            f')'
        )


def _gather_members(
    *,
    want_all: bool,
    indent: str,
    analyzer_order: dict[str, int],
    attr_docs: dict[tuple[str, str], list[str]],
    config: _AutodocConfig,
    current_document: _CurrentDocument,
    events: EventManager,
    get_attr: _AttrGetter,
    options: _AutoDocumenterOptions,
    parent_modname: str,
    props: _ItemProperties,
    ref_context: Mapping[str, str | None],
    reread_always: MutableSet[str],
) -> list[tuple[_ItemProperties, bool, str]]:
    """Generate reST for member documentation.

    If *want_all* is True, document all members, else those given by
    *self.options.members*.
    """
    if props.obj_type not in {'module', 'class', 'exception'}:
        msg = 'must be implemented in subclasses'
        raise NotImplementedError(msg)
    assert isinstance(props, (_ModuleProperties, _ClassDefProperties))

    indent += '   ' * (props.obj_type != 'module')

    # set current namespace for finding members
    current_document.autodoc_module = props.module_name
    if props.parts:
        current_document.autodoc_class = props.parts[0]

    inherited_members = frozenset(options.inherited_members or ())
    found_members = _get_members_to_document(
        want_all=want_all,
        get_attr=get_attr,
        class_signature=config.autodoc_class_signature,
        inherit_docstrings=config.autodoc_inherit_docstrings,
        props=props,
        opt_members=options.members or (),
        inherited_members=inherited_members,
        opt_private_members=options.private_members,
        opt_special_members=options.special_members,
        ignore_module_all=bool(options.ignore_module_all),
        attr_docs=attr_docs,
    )
    filtered_members = _filter_members(
        found_members,
        want_all=want_all,
        events=events,
        get_attr=get_attr,
        class_signature=config.autodoc_class_signature,
        inherit_docstrings=config.autodoc_inherit_docstrings,
        options=options,
        props=props,
        inherited_members=inherited_members,
        exclude_members=options.exclude_members,
        special_members=options.special_members,
        private_members=options.private_members,
        undoc_members=options.undoc_members,
        attr_docs=attr_docs,
    )
    # document non-skipped members
    member_documenters: list[tuple[_ItemProperties, bool, str]] = []
    for member_name, member, is_attr in filtered_members:
        # prefer the object type with the highest priority
        obj_type = _best_object_type_for_member(
            member=member,
            member_name=member_name,
            is_attr=is_attr,
            parent_obj_type=props.obj_type,
            parent_props=props,
        )
        if not obj_type:
            # don't know how to document this member
            continue
        # give explicitly separated module name, so that members
        # of inner classes can be documented
        dotted_parts = '.'.join((*props.parts, member_name))
        full_name = f'{props.module_name}::{dotted_parts}'

        # We now try to import all objects before ordering them. This is to
        # avoid possible circular imports if we were to import objects after
        # their associated documenters have been sorted.
        member_props = _load_object_by_name(
            name=full_name,
            objtype=obj_type,
            current_document=current_document,
            config=config,
            events=events,
            get_attr=get_attr,
            options=options,
            parent_modname=parent_modname,
            ref_context=ref_context,
            reread_always=reread_always,
        )
        if member_props is None:
            continue
        member_documenters.append((member_props, is_attr, indent))

    member_order = options.member_order or config.autodoc_member_order
    member_documenters = _sort_members(
        member_documenters,
        member_order,
        ignore_module_all=bool(options.ignore_module_all),
        analyzer_order=analyzer_order,
        props=props,
    )

    # reset current objects
    current_document.autodoc_module = ''
    current_document.autodoc_class = ''

    return member_documenters


def _get_members_to_document(
    *,
    want_all: bool,
    get_attr: _AttrGetter,
    class_signature: Literal['mixed', 'separated'],
    inherit_docstrings: bool,
    props: _ModuleProperties | _ClassDefProperties,
    opt_members: ALL_T | Sequence[str],
    inherited_members: Set[str],
    opt_private_members: ALL_T | Sequence[str] | None,
    opt_special_members: ALL_T | Sequence[str] | None,
    ignore_module_all: bool,
    attr_docs: dict[tuple[str, str], list[str]],
) -> list[ObjectMember]:
    """Find out which members are documentable

    If *want_all* is True, return all members.  Else, only return those
    members given by *self.options.members* (which may also be None).

    Filter the given member list.

    Members are skipped if

    - they are private (except if given explicitly or the private-members
      option is set)
    - they are special methods (except if given explicitly or the
      special-members option is set)
    - they are undocumented (except if the undoc-members option is set)

    The user can override the skipping decision by connecting to the
    ``autodoc-skip-member`` event.
    """
    wanted_members: ALL_T | Set[str]
    if want_all:
        if (
            props.obj_type == 'module'
            and not ignore_module_all
            and props.all is not None
        ):
            wanted_members = frozenset(props.all)
        else:
            wanted_members = ALL
    else:
        # specific members given
        assert opt_members is not ALL

        # Merge :private-members: and :special-members: into :members:
        combined_members = set(opt_members)
        if opt_private_members is not None and opt_private_members is not ALL:
            combined_members.update(opt_private_members)
        if opt_special_members is not None and opt_special_members is not ALL:
            combined_members.update(opt_special_members)
        if class_signature == 'separated' and props.obj_type in {'class', 'exception'}:
            combined_members |= {'__new__', '__init__'}  # show __init__() method
        wanted_members = frozenset(combined_members)

    object_members_map: dict[str, ObjectMember] = {}
    if props.obj_type == 'module':
        for name in dir(props._obj):
            try:
                value = safe_getattr(props._obj, name, None)
                if ismock(value):
                    value = undecorate(value)
                if name in wanted_members:
                    object_members_map[name] = ObjectMember(
                        name, value, docstring=attr_docs.get(('', name), [])
                    )
            except AttributeError:
                continue

        # annotation only member (e.g. attr: int)
        for name in inspect.getannotations(props._obj):
            if name not in object_members_map and name in wanted_members:
                object_members_map[name] = ObjectMember(
                    name, INSTANCE_ATTR, docstring=attr_docs.get(('', name), [])
                )

        obj_members_seq = list(object_members_map.values())
    elif props.obj_type in {'class', 'exception'}:
        # the members directly defined in the class
        obj_dict = get_attr(props._obj, '__dict__', {})

        # enum members
        if isenumclass(props._obj):
            for name, defining_class, value in _filter_enum_dict(
                props._obj, get_attr, obj_dict
            ):
                # the order of occurrence of *name* matches obj's MRO,
                # allowing inherited attributes to be shadowed correctly
                if unmangled := unmangle(defining_class, name):
                    if unmangled in wanted_members:
                        object_members_map[unmangled] = ObjectMember(
                            unmangled, value, class_=defining_class
                        )

        # members in __slots__
        try:
            subject___slots__ = getslots(props._obj)
            if subject___slots__:
                for name, subject_docstring in subject___slots__.items():
                    if name not in wanted_members:
                        continue
                    if isinstance(subject_docstring, str):
                        subject_doclines = subject_docstring.splitlines()
                    else:
                        subject_doclines = None
                    object_members_map[name] = ObjectMember(
                        name,
                        SLOTS_ATTR,
                        class_=props._obj,
                        docstring=subject_doclines,
                    )
        except (TypeError, ValueError):
            pass

        # other members
        for name in dir(props._obj):
            try:
                value = get_attr(props._obj, name)
                if ismock(value):
                    value = undecorate(value)

                unmangled = unmangle(props._obj, name)
                if (
                    unmangled
                    and unmangled not in object_members_map
                    and unmangled in wanted_members
                ):
                    if name in obj_dict:
                        object_members_map[unmangled] = ObjectMember(
                            unmangled, value, class_=props._obj
                        )
                    else:
                        object_members_map[unmangled] = ObjectMember(unmangled, value)
            except AttributeError:
                continue

        try:
            for cls in getmro(props._obj):
                try:
                    modname = safe_getattr(cls, '__module__')
                    qualname = safe_getattr(cls, '__qualname__')
                except AttributeError:
                    qualname = None
                    analyzer = None
                else:
                    try:
                        analyzer = ModuleAnalyzer.for_module(modname)
                        analyzer.analyze()
                    except PycodeError:
                        analyzer = None

                # annotation only member (ex. attr: int)
                for name in getannotations(cls):
                    unmangled = unmangle(cls, name)
                    if (
                        unmangled
                        and unmangled not in object_members_map
                        and unmangled in wanted_members
                    ):
                        if analyzer and (qualname, unmangled) in analyzer.attr_docs:
                            attr_docstring = analyzer.attr_docs[qualname, unmangled]
                        else:
                            attr_docstring = None
                        object_members_map[unmangled] = ObjectMember(
                            unmangled,
                            INSTANCE_ATTR,
                            class_=cls,
                            docstring=attr_docstring,
                        )

                # append or complete instance attributes (cf. self.attr1) if analyzer knows
                if analyzer:
                    for (ns, name), attr_docstring in analyzer.attr_docs.items():
                        if ns == qualname and name not in object_members_map:
                            # otherwise unknown instance attribute
                            if name in wanted_members:
                                object_members_map[name] = ObjectMember(
                                    name,
                                    INSTANCE_ATTR,
                                    class_=cls,
                                    docstring=attr_docstring,
                                )
                        elif (
                            ns == qualname
                            and attr_docstring
                            and not object_members_map[name].docstring
                        ):
                            if cls != props._obj and not inherit_docstrings:
                                # If we are in the MRO of the class and not the class itself,
                                # and we do not want to inherit docstrings, then skip setting
                                # the docstring below
                                continue
                            # attribute is already known, because dir(props._obj)
                            # enumerates it. But it has no docstring yet
                            object_members_map[name].docstring = attr_docstring
        except AttributeError:
            pass

        if want_all and not inherited_members:
            obj_members_seq = [
                m for m in object_members_map.values() if m.class_ == props._obj
            ]
        else:
            obj_members_seq = list(object_members_map.values())
    else:
        raise ValueError

    if not want_all and opt_members is not ALL:
        for name in opt_members:
            if name in object_members_map:
                continue
            msg = __(
                'attribute %s is listed in :members: but is missing '
                'as it was not found in object %r'
            )
            LOGGER.warning(msg, name, props._obj, type='autodoc')
    return obj_members_seq


def _filter_members(
    obj_members_seq: Iterable[ObjectMember],
    *,
    want_all: bool,
    events: EventManager,
    get_attr: _AttrGetter,
    options: _AutoDocumenterOptions,
    props: _ModuleProperties | _ClassDefProperties,
    class_signature: Literal['mixed', 'separated'],
    inherit_docstrings: bool,
    inherited_members: Set[str],
    exclude_members: EMPTY_T | Set[str] | None,
    special_members: ALL_T | Sequence[str] | None,
    private_members: ALL_T | Sequence[str] | None,
    undoc_members: Literal[True] | None,
    attr_docs: dict[tuple[str, str], list[str]],
) -> Iterator[tuple[str, Any, bool]]:
    # search for members in source code too
    namespace = props.dotted_parts  # will be empty for modules

    # process members and determine which to skip
    for obj in obj_members_seq:
        member_name = obj.__name__
        member_obj = obj.object
        has_attr_doc = (namespace, member_name) in attr_docs
        try:
            keep = _should_keep_member(
                member_name=member_name,
                member_obj=member_obj,
                member_docstring=obj.docstring,
                member_cls=obj.class_,
                get_attr=get_attr,
                has_attr_doc=has_attr_doc,
                class_signature=class_signature,
                inherit_docstrings=inherit_docstrings,
                inherited_members=inherited_members,
                parent=props._obj,
                want_all=want_all,
                exclude_members=exclude_members,
                special_members=special_members,
                private_members=private_members,
                undoc_members=undoc_members,
            )
        except Exception as exc:
            LOGGER.warning(
                __(
                    'autodoc: failed to determine %s.%s (%r) to be documented, '
                    'the following exception was raised:\n%s'
                ),
                props.full_name,
                member_name,
                member_obj,
                exc,
                type='autodoc',
            )
            keep = False

        # give the user a chance to decide whether this member
        # should be skipped
        if events is not None:
            # let extensions preprocess docstrings
            skip_member = events.emit_firstresult(
                'autodoc-skip-member',
                props.obj_type,
                member_name,
                member_obj,
                not keep,
                options,
            )
            if skip_member is not None:
                keep = not skip_member

        if keep:
            # if is_attr is True, the member is documented as an attribute
            is_attr = member_obj is INSTANCE_ATTR or has_attr_doc
            yield member_name, member_obj, is_attr


def _best_object_type_for_member(
    member: Any,
    member_name: str,
    is_attr: bool,
    *,
    parent_obj_type: str,
    parent_props: _ItemProperties | None,
) -> _AutodocObjType | None:
    """Return the best object type that supports documenting *member*."""
    filtered = []

    # Don't document submodules automatically: 'module' is never returned.

    try:
        if isinstance(member, type) and issubclass(member, BaseException):
            # priority must be higher than 'class'
            filtered.append((20, 'exception'))
    except TypeError as exc:
        # It's possible for a member to be considered a type, but fail
        # issubclass checks due to not being a class. For example:
        # https://github.com/sphinx-doc/sphinx/issues/11654#issuecomment-1696790436
        msg = f'Failed to discern if member {member} is a BaseException subclass.'
        raise ValueError(msg) from exc

    if isinstance(member, type) or (is_attr and isinstance(member, (NewType, TypeVar))):
        # priority must be higher than 'function', 'class', and 'attribute'
        # as NewType can be an attribute and is a class.
        filtered.append((15, 'class'))

    if parent_obj_type in {'class', 'exception'}:
        if inspect.isproperty(member):
            # priority must be higher than 'attribute'
            filtered.append((11, 'property'))

        # See _get_documenter() in autosummary, parent_props might be None.
        elif parent_props is not None:
            # Support for class properties. Note: these only work on Python 3.9.
            __dict__ = safe_getattr(parent_props._obj, '__dict__', {})
            obj = __dict__.get(member_name)
            if isinstance(obj, classmethod) and inspect.isproperty(obj.__func__):
                # priority must be higher than 'attribute'
                filtered.append((11, 'property'))

    if parent_obj_type != 'module':
        if inspect.isattributedescriptor(member) or not (
            inspect.isroutine(member) or isinstance(member, type)
        ):
            # priority must be higher than 'method', else it will recognise
            # some non-data descriptors as methods
            filtered.append((10, 'attribute'))

    if inspect.isroutine(member) and parent_obj_type != 'module':
        # priority must be higher than 'function'
        filtered.append((1, 'method'))

    if (
        inspect.isfunction(member)
        or inspect.isbuiltin(member)
        or (inspect.isroutine(member) and parent_obj_type == 'module')
    ):
        # supports functions, builtins and bound methods exported
        # at the module level
        filtered.extend(((0, 'function'), (-1, 'decorator')))

    if isinstance(member, AnyTypeAliasType):
        filtered.append((0, 'type'))

    if parent_obj_type == 'module' and is_attr:
        filtered.append((-10, 'data'))

    if filtered:
        # return the highest priority object type
        return max(filtered, key=operator.itemgetter(0))[1]  # type: ignore[return-value]
    return None


def _sort_members(
    documenters: list[tuple[_ItemProperties, bool, str]],
    order: Literal['alphabetical', 'bysource', 'groupwise'],
    *,
    ignore_module_all: bool,
    analyzer_order: dict[str, int],
    props: _ItemProperties,
) -> list[tuple[_ItemProperties, bool, str]]:
    """Sort the given member list."""
    if order == 'groupwise':
        # sort by group; alphabetically within groups
        def group_order(entry: tuple[_ItemProperties, bool, str]) -> tuple[int, str]:
            return entry[0]._groupwise_order_key, entry[0].full_name

        documenters.sort(key=group_order)
    elif order == 'bysource':
        if (
            isinstance(props, _ModuleProperties)
            and not ignore_module_all
            and (module_all := props.all)
        ):
            # Sort by __all__
            module_all_idx = {name: idx for idx, name in enumerate(module_all)}
            module_all_len = len(module_all)

            def source_order(entry: tuple[_ItemProperties, bool, str]) -> int:
                fullname = entry[0].dotted_parts
                return module_all_idx.get(fullname, module_all_len)

            documenters.sort(key=source_order)

        # By default, member discovery order matches source order,
        # as dicts are insertion-ordered from Python 3.7.
        elif analyzer_order:
            # sort by source order, by virtue of the module analyzer
            order_len = len(analyzer_order)

            def source_order(entry: tuple[_ItemProperties, bool, str]) -> int:
                fullname = entry[0].dotted_parts
                return analyzer_order.get(fullname, order_len)

            documenters.sort(key=source_order)
    else:  # alphabetical
        documenters.sort(key=lambda entry: entry[0].full_name)

    return documenters


def unmangle(subject: Any, name: str) -> str | None:
    """Unmangle the given name."""
    try:
        if isclass(subject) and not name.endswith('__'):
            prefix = f'_{subject.__name__}__'
            if name.startswith(prefix):
                return name.replace(prefix, '__', 1)
            else:
                for cls in subject.__mro__:
                    prefix = f'_{cls.__name__}__'
                    if name.startswith(prefix):
                        # mangled attribute defined in parent class
                        return None
    except AttributeError:
        pass

    return name


def _filter_enum_dict(
    enum_class: type[Enum],
    attrgetter: _AttrGetter,
    enum_class_dict: Mapping[str, object],
) -> Iterator[tuple[str, type, Any]]:
    """Find the attributes to document of an enumeration class.

    The output consists of triplets ``(attribute name, defining class, value)``
    where the attribute name can appear more than once during the iteration
    but with different defining class. The order of occurrence is guided by
    the MRO of *enum_class*.
    """
    # attributes that were found on a mixin type or the data type
    candidate_in_mro: set[str] = set()
    # sunder names that were picked up (and thereby allowed to be redefined)
    # see: https://docs.python.org/3/howto/enum.html#supported-dunder-names
    sunder_names = {
        '_name_',
        '_value_',
        '_missing_',
        '_order_',
        '_generate_next_value_',
    }
    # attributes that can be picked up on a mixin type or the enum's data type
    public_names = {'name', 'value', *object.__dict__, *sunder_names}
    # names that are ignored by default
    ignore_names = Enum.__dict__.keys() - public_names

    def should_ignore(name: str, value: Any) -> bool:
        if name in sunder_names:
            return _is_native_enum_api(value, name)
        return name in ignore_names

    sentinel = object()

    def query(name: str, defining_class: type) -> tuple[str, type, Any] | None:
        value = attrgetter(enum_class, name, sentinel)
        if value is not sentinel:
            return name, defining_class, value
        return None

    # attributes defined on a parent type, possibly shadowed later by
    # the attributes defined directly inside the enumeration class
    for parent in enum_class.__mro__:
        if parent in {enum_class, Enum, object}:
            continue

        parent_dict = attrgetter(parent, '__dict__', {})
        for name, value in parent_dict.items():
            if should_ignore(name, value):
                continue

            candidate_in_mro.add(name)
            if (item := query(name, parent)) is not None:
                yield item

    # exclude members coming from the native Enum unless
    # they were redefined on a mixin type or the data type
    excluded_members = Enum.__dict__.keys() - candidate_in_mro
    yield from filter(
        None,
        (
            query(name, enum_class)
            for name in enum_class_dict
            if name not in excluded_members
        ),
    )

    # check if allowed members from ``Enum`` were redefined at the enum level
    special_names = sunder_names | public_names
    special_names &= enum_class_dict.keys()
    special_names &= Enum.__dict__.keys()
    for name in special_names:
        if (
            not _is_native_enum_api(enum_class_dict[name], name)
            and (item := query(name, enum_class)) is not None
        ):
            yield item


def _is_native_enum_api(obj: object, name: str) -> bool:
    """Check whether *obj* is the same as ``Enum.__dict__[name]``."""
    return unwrap_all(obj) is unwrap_all(Enum.__dict__[name])


def _should_keep_member(
    *,
    member_name: str,
    member_obj: Any,
    member_docstring: Sequence[str] | None,
    member_cls: Any,
    get_attr: _AttrGetter,
    has_attr_doc: bool,
    class_signature: Literal['mixed', 'separated'],
    inherit_docstrings: bool,
    inherited_members: Set[str],
    parent: Any,
    want_all: bool,
    exclude_members: EMPTY_T | Set[str] | None,
    special_members: ALL_T | Sequence[str] | None,
    private_members: ALL_T | Sequence[str] | None,
    undoc_members: Literal[True] | None,
) -> bool:
    if member_docstring:
        # hack for ClassDocumenter to inject docstring
        doclines: Sequence[str] | None = member_docstring
    else:
        doc = getdoc(
            member_obj,
            get_attr,
            inherit_docstrings,
            parent,
            member_name,
        )
        # Ignore non-string __doc__
        doclines = doc.splitlines() if isinstance(doc, str) else None

        # if the member __doc__ is the same as self's __doc__, it's just
        # inherited and therefore not the member's doc
        cls = get_attr(member_obj, '__class__', None)
        if cls:
            cls_doc = get_attr(cls, '__doc__', None)
            if cls_doc == doc:
                doclines = None

    if doclines is not None:
        doc, metadata = separate_metadata('\n'.join(doclines))
    else:
        doc = ''
        metadata = {}
    has_doc = bool(doc or undoc_members)

    if 'private' in metadata:
        # consider a member private if docstring has "private" metadata
        is_private = True
    elif 'public' in metadata:
        # consider a member public if docstring has "public" metadata
        is_private = False
    else:
        is_private = member_name.startswith('_')

    if ismock(member_obj) and not has_attr_doc:
        # mocked module or object
        return False

    if exclude_members and member_name in exclude_members:
        # remove members given by exclude-members
        return False

    if not want_all:
        # keep documented attributes
        return has_doc or has_attr_doc

    is_filtered_inherited_member = _is_filtered_inherited_member(
        member_name,
        member_cls=member_cls,
        parent=parent,
        inherited_members=inherited_members,
        get_attr=get_attr,
    )

    if special_member_re.match(member_name):
        # special __methods__
        if member_name == '__doc__' or is_filtered_inherited_member:
            return False
        if special_members and member_name in special_members:
            return has_doc
        if (
            class_signature == 'separated'
            and member_name in {'__new__', '__init__'}
            and inspect.isclass(parent)
        ):
            return has_doc  # show __init__() method
        return False

    if is_private:
        if has_attr_doc or has_doc:
            if private_members is None:  # NoQA: SIM114
                return False
            elif has_doc and is_filtered_inherited_member:
                return False
            return member_name in private_members
        return False

    if has_attr_doc:
        # keep documented attributes
        return True

    if is_filtered_inherited_member:
        return False

    # ignore undocumented members if :undoc-members: is not given
    return has_doc


def _is_filtered_inherited_member(
    member_name: str,
    *,
    member_cls: Any,
    parent: Any,
    inherited_members: Set[str],
    get_attr: _AttrGetter,
) -> bool:
    if not inspect.isclass(parent):
        return False

    seen = set()
    for cls in parent.__mro__:
        if member_name in cls.__dict__:
            seen.add(cls)
        if (
            cls.__name__ in inherited_members
            and cls != parent
            and any(issubclass(potential_child, cls) for potential_child in seen)
        ):
            # given member is a member of specified *super class*
            return True
        if member_cls is cls:
            return False
        if member_name in cls.__dict__:
            return False
        if member_name in get_attr(cls, '__annotations__', {}):
            return False
    return False



"""Domain indices."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, NamedTuple

from sphinx.errors import SphinxError

if TYPE_CHECKING:
    from collections.abc import Iterable
    from typing import ClassVar

    from sphinx.domains import Domain


class IndexEntry(NamedTuple):
    """An index entry.

    .. note::

       The *qualifier* and *description* are not rendered for some output formats,
       such as LaTeX.
    """

    #: The name of the index entry to be displayed.
    name: str

    #: The sub-entry related type. One of:
    #:
    #: ``0``
    #:   A normal entry.
    #: ``1``
    #:   An entry with sub-entries.
    #: ``2``
    #:   A sub-entry.
    subtype: int

    #: *docname* where the entry is located.
    docname: str

    #: Anchor for the entry within `docname`
    anchor: str

    #: Extra info for the entry.
    extra: str

    #: Qualifier for the description.
    qualifier: str

    #: Description for the entry.
    descr: str


class Index(ABC):
    """An Index is the description for a domain-specific index.  To add an index to
    a domain, subclass Index, overriding the three name attributes:

    * `name` is an identifier used for generating file names.
      It is also used for a hyperlink target for the index. Therefore, users can
      refer the index page using ``ref`` role and a string which is combined
      domain name and ``name`` attribute (ex. ``:ref:`py-modindex```).
    * `localname` is the section title for the index.
    * `shortname` is a short name for the index, for use in the relation bar in
      HTML output.  Can be empty to disable entries in the relation bar.

    and providing a :meth:`generate` method.  Then, add the index class to
    your domain's `indices` list.  Extensions can add indices to existing
    domains using :meth:`~sphinx.application.Sphinx.add_index_to_domain`.

    .. versionchanged:: 3.0

       Index pages can be referred by domain name and index name via
       :rst:role:`ref` role.
    """

    name: ClassVar[str]
    localname: ClassVar[str]
    shortname: ClassVar[str | None] = None

    def __init__(self, domain: Domain) -> None:
        if not self.name or self.localname is None:
            msg = f'Index subclass {self.__class__.__name__} has no valid name or localname'
            raise SphinxError(msg)
        self.domain = domain

    @abstractmethod
    def generate(
        self, docnames: Iterable[str] | None = None
    ) -> tuple[list[tuple[str, list[IndexEntry]]], bool]:
        """Get entries for the index.

        If ``docnames`` is given, restrict to entries referring to these
        docnames.

        The return value is a tuple of ``(content, collapse)``:

        ``collapse``
          A boolean that determines if sub-entries should start collapsed (for
          output formats that support collapsing sub-entries).

        ``content``:
          A sequence of ``(letter, entries)`` tuples, where ``letter`` is the
          "heading" for the given ``entries``, usually the starting letter, and
          ``entries`` is a sequence of single entries.
          Each entry is an :py:class:`IndexEntry`.
        """
        raise NotImplementedError



"""Serialise objects to a stable representation."""

from __future__ import annotations

import hashlib
import json
import types
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def stable_hash(obj: Any) -> str:
    """Return a stable hash for a Python data structure.

    We can't just use the md5 of str(obj) as the order of collections
    may be random.
    """
    if isinstance(obj, dict):
        obj = sorted(map(stable_hash, obj.items()))
    if isinstance(obj, (list, tuple, set, frozenset)):
        obj = sorted(map(stable_hash, obj))
    elif isinstance(obj, (type, types.FunctionType)):
        # The default repr() of functions includes the ID, which is not ideal.
        # We use the fully qualified name instead.
        obj = f'{obj.__module__}.{obj.__qualname__}'
    return hashlib.md5(str(obj).encode(), usedforsecurity=False).hexdigest()


def stable_str(obj: Any, *, indent: int | None = None) -> str:
    """Return a stable string representation of a Python data structure.

    We can't just use ``str(obj)`` as the order of collections may be random.
    """
    return json.dumps(_stable_str_prep(obj), indent=indent)


def _stable_str_prep(obj: Any) -> dict[str, Any] | list[Any] | str:
    if isinstance(obj, dict):
        # Convert to a sorted dict
        lst: list[tuple[Any, Any]] = [
            (_stable_str_prep(k), _stable_str_prep(v)) for k, v in obj.items()
        ]
        lst.sort()
        return dict(lst)
    if isinstance(obj, (list, tuple, set, frozenset)):
        # Convert to a sorted list
        return sorted(map(_stable_str_prep, obj), key=str)
    if isinstance(obj, (type, types.FunctionType)):
        # The default repr() of functions includes the ID, which is not ideal.
        # We use the fully qualified name instead.
        return f'{obj.__module__}.{obj.__qualname__}'
    # We can't do any better, just use the string representation
    return str(obj)



"""Importer utilities for autodoc"""

from __future__ import annotations

import contextlib
import importlib
import os
import sys
import traceback
import typing
from importlib.abc import FileLoader
from importlib.machinery import EXTENSION_SUFFIXES
from importlib.util import decode_source, find_spec, module_from_spec, spec_from_loader
from pathlib import Path
from typing import TYPE_CHECKING

from sphinx.errors import PycodeError
from sphinx.ext.autodoc._dynamic._mock import ismock, mock, undecorate
from sphinx.ext.autodoc._sentinels import RUNTIME_INSTANCE_ATTRIBUTE, UNINITIALIZED_ATTR
from sphinx.ext.autodoc._shared import LOGGER
from sphinx.pycode import ModuleAnalyzer
from sphinx.util import inspect
from sphinx.util.inspect import isclass, safe_getattr
from sphinx.util.typing import get_type_hints

if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from importlib.machinery import ModuleSpec
    from types import ModuleType
    from typing import Any, Protocol

    from sphinx.ext.autodoc._property_types import _AutodocObjType

    class _AttrGetter(Protocol):
        def __call__(self, obj: Any, name: str, default: Any = ..., /) -> Any: ...


_NATIVE_SUFFIXES: frozenset[str] = frozenset({'.pyx', *EXTENSION_SUFFIXES})


class _ImportedObject:
    #: module containing the object to document
    module: ModuleType | None

    #: parent/owner of the object to document
    parent: Any

    #: name of the object to document
    object_name: str

    #: object to document
    obj: Any

    def __init__(
        self,
        *,
        module: ModuleType | None = None,
        parent: Any,
        object_name: str = '',
        obj: Any,
    ) -> None:
        self.module = module
        self.parent = parent
        self.object_name = object_name
        self.obj = obj

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} {self.__dict__}>'


def _import_object(
    *,
    get_attr: _AttrGetter = safe_getattr,
    mock_imports: Sequence[str],
    module_name: str,
    obj_path: Sequence[str],
    obj_type: _AutodocObjType,
    type_aliases: Mapping[str, str] | None,
) -> _ImportedObject | None:
    """Import the module and get the object to document."""
    try:
        with mock(mock_imports):
            im = _import_from_module_and_path(
                module_name=module_name, obj_path=obj_path, get_attr=get_attr
            )
    except ImportError as exc:
        if obj_type == 'data':
            im_ = _import_data_declaration(
                module_name=module_name,
                obj_path=obj_path,
                mock_imports=mock_imports,
                type_aliases=type_aliases,
            )
            if im_ is not None:
                return im_
        elif obj_type == 'attribute':
            im_ = _import_attribute_declaration(
                module_name=module_name,
                obj_path=obj_path,
                mock_imports=mock_imports,
                type_aliases=type_aliases,
                get_attr=get_attr,
            )
            if im_ is not None:
                return im_
        LOGGER.warning(exc.args[0], type='autodoc', subtype='import_object')
        return None

    if ismock(im.obj):
        im.obj = undecorate(im.obj)
    return im


def _import_from_module_and_path(
    *,
    module_name: str,
    obj_path: Sequence[str],
    get_attr: _AttrGetter = safe_getattr,
) -> _ImportedObject:
    obj_path = list(obj_path)
    if obj_path:
        LOGGER.debug('[autodoc] from %s import %s', module_name, '.'.join(obj_path))
    else:
        LOGGER.debug('[autodoc] import %s', module_name)

    module = None
    exc_on_importing = None
    try:
        while module is None:
            try:
                module = _import_module(module_name, try_reload=True)
                LOGGER.debug('[autodoc] import %s => %r', module_name, module)
            except ImportError as exc:
                LOGGER.debug('[autodoc] import %s => failed', module_name)
                exc_on_importing = exc
                if '.' not in module_name:
                    raise

                # retry with parent module
                module_name, _, name = module_name.rpartition('.')
                obj_path.insert(0, name)

        obj = module
        parent = None
        object_name = ''
        for attr_name in obj_path:
            parent = obj
            LOGGER.debug('[autodoc] getattr(_, %r)', attr_name)
            mangled_name = _mangle_name(obj, attr_name)
            obj = get_attr(obj, mangled_name)

            try:
                LOGGER.debug('[autodoc] => %r', obj)
            except TypeError:
                # fallback of failure on logging for broken object
                # See: https://github.com/sphinx-doc/sphinx/issues/9095
                LOGGER.debug('[autodoc] => %r', (obj,))

            object_name = attr_name
        return _ImportedObject(
            module=module,
            parent=parent,
            object_name=object_name,
            obj=obj,
        )
    except (AttributeError, ImportError) as exc:
        if isinstance(exc, AttributeError) and exc_on_importing:
            # restore ImportError
            exc = exc_on_importing

        if obj_path:
            dotted_objpath = '.'.join(obj_path)
            err_parts = [
                f'autodoc: failed to import {dotted_objpath!r} '
                f'from module {module_name!r}'
            ]
        else:
            err_parts = [f'autodoc: failed to import {module_name!r}']

        if isinstance(exc, ImportError):
            # _import_module() raises ImportError having real exception obj and
            # traceback
            real_exc = exc.args[0]
            traceback_msg = ''.join(traceback.format_exception(exc))
            if isinstance(real_exc, SystemExit):
                err_parts.append(
                    'the module executes module level statement '
                    'and it might call sys.exit().'
                )
            elif isinstance(real_exc, ImportError) and real_exc.args:
                err_parts.append(
                    f'the following exception was raised:\n{real_exc.args[0]}'
                )
            else:
                err_parts.append(
                    f'the following exception was raised:\n{traceback_msg}'
                )
        else:
            err_parts.append(
                f'the following exception was raised:\n{traceback.format_exc()}'
            )

        errmsg = '; '.join(err_parts)
        LOGGER.debug(errmsg)
        raise ImportError(errmsg) from exc


def _import_module(modname: str, try_reload: bool = False) -> Any:
    if modname in sys.modules:
        return sys.modules[modname]

    skip_pyi = bool(os.getenv('SPHINX_AUTODOC_IGNORE_NATIVE_MODULE_TYPE_STUBS', ''))
    original_module_names = frozenset(sys.modules)
    try:
        spec = find_spec(modname)
        if spec is None:
            msg = f'No module named {modname!r}'
            raise ModuleNotFoundError(msg, name=modname)  # NoQA: TRY301
        spec, pyi_path = _find_type_stub_spec(spec, modname)
        if skip_pyi or pyi_path is None:
            module = importlib.import_module(modname)
        else:
            if spec.loader is None:
                msg = 'missing loader'
                raise ImportError(msg, name=spec.name)  # NoQA: TRY301
            sys.modules[modname] = module = module_from_spec(spec)
            spec.loader.exec_module(module)
    except ImportError:
        raise
    except BaseException as exc:
        # Importing modules may cause any side effects, including
        # SystemExit, so we need to catch all errors.
        raise ImportError(exc, traceback.format_exc()) from exc
    if try_reload and os.environ.get('SPHINX_AUTODOC_RELOAD_MODULES'):
        new_modules = [m for m in sys.modules if m not in original_module_names]
        # Try reloading modules with ``typing.TYPE_CHECKING == True``.
        try:
            typing.TYPE_CHECKING = True  # type: ignore[misc]
            # Ignore failures; we've already successfully loaded these modules
            with contextlib.suppress(ImportError, KeyError):
                for m in new_modules:
                    mod_path = getattr(sys.modules[m], '__file__', '')
                    if mod_path and mod_path.endswith('.pyi'):
                        continue
                    _reload_module(sys.modules[m])
        finally:
            typing.TYPE_CHECKING = False  # type: ignore[misc]
        module = sys.modules[modname]
    return module


def _find_type_stub_spec(
    spec: ModuleSpec, modname: str
) -> tuple[ModuleSpec, Path | None]:
    """Try finding a spec for a PEP 561 '.pyi' stub file for native modules."""
    if spec.origin is None:
        return spec, None

    for suffix in _NATIVE_SUFFIXES:
        if not spec.origin.endswith(suffix):
            continue
        pyi_path = Path(spec.origin.removesuffix(suffix) + '.pyi')
        if not pyi_path.is_file():
            continue
        pyi_loader = _StubFileLoader(modname, path=str(pyi_path))
        pyi_spec = spec_from_loader(modname, loader=pyi_loader)
        if pyi_spec is not None:
            return pyi_spec, pyi_path
    return spec, None


class _StubFileLoader(FileLoader):
    """Load modules from ``.pyi`` stub files."""

    def get_source(self, fullname: str) -> str:
        path = self.get_filename(fullname)
        for suffix in _NATIVE_SUFFIXES:
            if not path.endswith(suffix):
                continue
            path = path.removesuffix(suffix) + '.pyi'
        try:
            source_bytes = self.get_data(path)
        except OSError as exc:
            raise ImportError from exc
        return decode_source(source_bytes)


def _reload_module(module: ModuleType) -> Any:
    """Call importlib.reload(module), convert exceptions to ImportError"""
    try:
        return importlib.reload(module)
    except BaseException as exc:
        # Importing modules may cause any side effects, including
        # SystemExit, so we need to catch all errors.
        raise ImportError(exc, traceback.format_exc()) from exc


def _mangle_name(subject: Any, name: str) -> str:
    """Mangle the given name."""
    try:
        if isclass(subject) and name.startswith('__') and not name.endswith('__'):
            return f'_{subject.__name__}{name}'
    except AttributeError:
        pass

    return name


def _import_data_declaration(
    *,
    module_name: str,
    obj_path: Sequence[str],
    mock_imports: Sequence[str],
    type_aliases: Mapping[str, str] | None,
) -> _ImportedObject | None:
    # annotation only instance variable (PEP-526)
    try:
        with mock(mock_imports):
            parent = _import_module(module_name)
        annotations = get_type_hints(parent, None, type_aliases, include_extras=True)
        if obj_path[-1] in annotations:
            im = _ImportedObject(
                parent=parent,
                obj=UNINITIALIZED_ATTR,
            )
            return im
    except ImportError:
        pass
    return None


def _import_attribute_declaration(
    *,
    module_name: str,
    obj_path: Sequence[str],
    mock_imports: Sequence[str],
    type_aliases: Mapping[str, str] | None,
    get_attr: _AttrGetter = safe_getattr,
) -> _ImportedObject | None:
    # Support runtime & uninitialized instance attributes.
    #
    # The former are defined in __init__() methods with doc-comments.
    # The latter are PEP-526 style annotation only annotations.
    #
    # class Foo:
    #     attr: int  #: uninitialized attribute
    #
    #     def __init__(self):
    #         self.attr = None  #: runtime attribute
    try:
        with mock(mock_imports):
            ret = _import_from_module_and_path(
                module_name=module_name, obj_path=obj_path[:-1], get_attr=get_attr
            )
        parent = ret.obj
        if _is_runtime_instance_attribute(parent=parent, obj_path=obj_path):
            im = _ImportedObject(
                parent=parent,
                obj=RUNTIME_INSTANCE_ATTRIBUTE,
            )
            return im
        elif _is_uninitialized_instance_attribute(
            parent=parent, obj_path=obj_path, type_aliases=type_aliases
        ):
            im = _ImportedObject(
                parent=parent,
                obj=UNINITIALIZED_ATTR,
            )
            return im
    except ImportError:
        pass
    return None


def _is_runtime_instance_attribute(*, parent: Any, obj_path: Sequence[str]) -> bool:
    """Check the subject is an attribute defined in __init__()."""
    # An instance variable defined in __init__().
    if _get_attribute_comment(parent=parent, obj_path=obj_path, attrname=obj_path[-1]):
        return True
    return _is_runtime_instance_attribute_not_commented(
        parent=parent, obj_path=obj_path
    )


def _is_runtime_instance_attribute_not_commented(
    *, parent: Any, obj_path: Sequence[str]
) -> bool:
    """Check the subject is an attribute defined in __init__() without comment."""
    for cls in inspect.getmro(parent):
        try:
            module = safe_getattr(cls, '__module__')
            qualname = safe_getattr(cls, '__qualname__')

            analyzer = ModuleAnalyzer.for_module(module)
            analyzer.analyze()
            if qualname and obj_path:
                key = f'{qualname}.{obj_path[-1]}'
                if key in analyzer.tagorder:
                    return True
        except (AttributeError, PycodeError):
            pass

    return False


def _get_attribute_comment(
    parent: Any, obj_path: Sequence[str], attrname: str
) -> list[str] | None:
    for cls in inspect.getmro(parent):
        try:
            module = safe_getattr(cls, '__module__')
            qualname = safe_getattr(cls, '__qualname__')

            analyzer = ModuleAnalyzer.for_module(module)
            analyzer.analyze()
            if qualname and obj_path:
                key = (qualname, attrname)
                if key in analyzer.attr_docs:
                    return list(analyzer.attr_docs[key])
        except (AttributeError, PycodeError):
            pass

    return None


def _is_uninitialized_instance_attribute(
    *, parent: Any, obj_path: Sequence[str], type_aliases: Mapping[str, str] | None
) -> bool:
    """Check the subject is an annotation only attribute."""
    annotations = get_type_hints(parent, None, type_aliases, include_extras=True)
    return obj_path[-1] in annotations



"""This module contains code shared between intersphinx modules."""

from __future__ import annotations

from typing import TYPE_CHECKING

from sphinx.util import logging

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any, Final, NoReturn

    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import Inventory

    #: The inventory project URL to which links are resolved.
    #:
    #: This value is unique in :confval:`intersphinx_mapping`.
    InventoryURI = str

    #: The inventory (non-empty) name.
    #:
    #: It is unique and in bijection with an inventory remote URL.
    InventoryName = str

    #: A target (local or remote) containing the inventory data to fetch.
    #:
    #: Empty strings are not expected and ``None`` indicates the default
    #: inventory file name :data:`~sphinx.builder.html.INVENTORY_FILENAME`.
    InventoryLocation = str | None

    #: Inventory cache entry. The integer field is the cache expiration time.
    type InventoryCacheEntry = tuple[InventoryName, int, Inventory]

    #: The type of :confval:`intersphinx_mapping` *after* normalisation.
    IntersphinxMapping = dict[
        InventoryName,
        tuple[InventoryName, tuple[InventoryURI, tuple[InventoryLocation, ...]]],
    ]

LOGGER: Final[logging.SphinxLoggerAdapter] = logging.getLogger('sphinx.ext.intersphinx')


class _IntersphinxProject:
    name: InventoryName
    target_uri: InventoryURI
    locations: tuple[InventoryLocation, ...]

    __slots__ = {
        'name':       'The inventory name. '
                      'It is unique and in bijection with an remote inventory URL.',
        'target_uri': 'The inventory project URL to which links are resolved. '
                      'It is unique and in bijection with an inventory name.',
        'locations':  'A tuple of local or remote targets containing '
                      'the inventory data to fetch. '
                      'None indicates the default inventory file name.',
    }  # fmt: skip

    def __init__(
        self,
        *,
        name: InventoryName,
        target_uri: InventoryURI,
        locations: Sequence[InventoryLocation],
    ) -> None:
        if not name or not isinstance(name, str):
            msg = 'name must be a non-empty string'
            raise ValueError(msg)
        if not target_uri or not isinstance(target_uri, str):
            msg = 'target_uri must be a non-empty string'
            raise ValueError(msg)
        if not locations or not isinstance(locations, tuple):
            msg = 'locations must be a non-empty tuple'
            raise ValueError(msg)
        if any(
            location is not None and (not location or not isinstance(location, str))
            for location in locations
        ):
            msg = 'locations must be a tuple of strings or None'
            raise ValueError(msg)
        object.__setattr__(self, 'name', name)
        object.__setattr__(self, 'target_uri', target_uri)
        object.__setattr__(self, 'locations', tuple(locations))

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}('
            f'name={self.name!r}, '
            f'target_uri={self.target_uri!r}, '
            f'locations={self.locations!r})'
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _IntersphinxProject):
            return NotImplemented
        return (
            self.name == other.name
            and self.target_uri == other.target_uri
            and self.locations == other.locations
        )

    def __hash__(self) -> int:
        return hash((self.name, self.target_uri, self.locations))

    def __setattr__(self, key: str, value: Any) -> NoReturn:
        msg = f'{self.__class__.__name__} is immutable'
        raise AttributeError(msg)

    def __delattr__(self, key: str) -> NoReturn:
        msg = f'{self.__class__.__name__} is immutable'
        raise AttributeError(msg)


class InventoryAdapter:
    """Inventory adapter for environment"""

    def __init__(self, env: BuildEnvironment) -> None:
        self.env = env

        if not hasattr(env, 'intersphinx_cache'):
            # initial storage when fetching inventories before processing
            self.env.intersphinx_cache = {}  # type: ignore[attr-defined]

            self.env.intersphinx_inventory = {}  # type: ignore[attr-defined]
            self.env.intersphinx_named_inventory = {}  # type: ignore[attr-defined]

    @property
    def cache(self) -> dict[InventoryURI, InventoryCacheEntry]:
        """Intersphinx cache.

        - Key is the URI of the remote inventory.
        - Element one is the key given in the Sphinx :confval:`intersphinx_mapping`.
        - Element two is a time value for cache invalidation, an integer.
        - Element three is the loaded remote inventory of type :class:`!Inventory`.
        """
        return self.env.intersphinx_cache  # type: ignore[attr-defined]

    @property
    def main_inventory(self) -> Inventory:
        return self.env.intersphinx_inventory  # type: ignore[attr-defined]

    @property
    def named_inventory(self) -> dict[InventoryName, Inventory]:
        return self.env.intersphinx_named_inventory  # type: ignore[attr-defined]

    def clear(self) -> None:
        self.env.intersphinx_inventory.clear()  # type: ignore[attr-defined]
        self.env.intersphinx_named_inventory.clear()  # type: ignore[attr-defined]



from __future__ import annotations

from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Mapping, Set
    from typing import Any, Final, Literal, NoReturn, Self

    from docutils import nodes

    from sphinx.domains import Domain
    from sphinx.domains.c import CDomain
    from sphinx.domains.changeset import ChangeSetDomain
    from sphinx.domains.citation import CitationDomain
    from sphinx.domains.cpp import CPPDomain
    from sphinx.domains.index import IndexDomain
    from sphinx.domains.javascript import JavaScriptDomain
    from sphinx.domains.math import MathDomain
    from sphinx.domains.python import PythonDomain
    from sphinx.domains.rst import ReSTDomain
    from sphinx.domains.std import StandardDomain
    from sphinx.environment import BuildEnvironment
    from sphinx.ext.duration import DurationDomain
    from sphinx.ext.todo import TodoDomain
    from sphinx.registry import SphinxComponentRegistry


class _DomainsContainer:
    """Container for domain instances.

    This class is private, including its name, constructor, and all methods.
    Any or all of these will change without notice or warning in any release.

    The public interface is restricted to:

    * the``domains.['<domain-name>']`` mapping interface
    * the ``domains.<core-domain-name>`` attributes for core domains.
    * the `.get()``, ``.keys()``, ``.items()``, and ``.values()`` methods.

    Additionally, this class supports ``iter`` and ``len``,
    and provides membership testing via the ``in`` operator.
    """

    __slots__ = (
        '_domain_instances',
        'c_domain',
        'changeset_domain',
        'citation_domain',
        'cpp_domain',
        'index_domain',
        'javascript_domain',
        'math_domain',
        'python_domain',
        'restructuredtext_domain',
        'standard_domain',
    )

    #: First-party domains in :mod:`sphinx.domains`
    _core_domains: Final = frozenset({
        'std',
        # Language-specific domains
        'c',
        'cpp',
        'js',
        'py',
        'rst',
        # Other core domains
        'changeset',
        'citation',
        'index',
        'math',
    })

    @classmethod
    def _from_environment(
        cls, env: BuildEnvironment, /, *, registry: SphinxComponentRegistry
    ) -> Self:
        create_domains = registry.create_domains
        # Initialise domains
        if domains := {domain.name: domain for domain in create_domains(env)}:
            return cls(**domains)  # type: ignore[arg-type]

        return cls._from_environment_default(env=env)

    @classmethod
    def _from_environment_default(cls, *, env: BuildEnvironment) -> Self:
        """Return a default instance with every domain we require."""
        from sphinx.domains.c import CDomain
        from sphinx.domains.changeset import ChangeSetDomain
        from sphinx.domains.citation import CitationDomain
        from sphinx.domains.cpp import CPPDomain
        from sphinx.domains.index import IndexDomain
        from sphinx.domains.javascript import JavaScriptDomain
        from sphinx.domains.math import MathDomain
        from sphinx.domains.python import PythonDomain
        from sphinx.domains.rst import ReSTDomain
        from sphinx.domains.std import StandardDomain

        return cls(
            c=CDomain(env),
            changeset=ChangeSetDomain(env),
            citation=CitationDomain(env),
            cpp=CPPDomain(env),
            index=IndexDomain(env),
            js=JavaScriptDomain(env),
            math=MathDomain(env),
            py=PythonDomain(env),
            rst=ReSTDomain(env),
            std=StandardDomain(env),
        )

    def __init__(
        self,
        *,
        c: CDomain,
        cpp: CPPDomain,
        js: JavaScriptDomain,
        py: PythonDomain,
        rst: ReSTDomain,
        std: StandardDomain,
        changeset: ChangeSetDomain,
        citation: CitationDomain,
        index: IndexDomain,
        math: MathDomain,
        **domains: Domain,
    ) -> None:
        # All domains, including core.
        # Implemented as a dict for backwards compatibility.
        self._domain_instances: Mapping[str, Domain] = {
            'c': c,
            'changeset': changeset,
            'citation': citation,
            'cpp': cpp,
            'index': index,
            'js': js,
            'math': math,
            'py': py,
            'rst': rst,
            'std': std,
            **domains,
        }

        # Provide typed attributes for the core domains
        self.standard_domain: StandardDomain = std
        self.c_domain: CDomain = c
        self.cpp_domain: CPPDomain = cpp
        self.javascript_domain: JavaScriptDomain = js
        self.python_domain: PythonDomain = py
        self.restructuredtext_domain: ReSTDomain = rst
        self.changeset_domain: ChangeSetDomain = changeset
        self.citation_domain: CitationDomain = citation
        self.index_domain: IndexDomain = index
        self.math_domain: MathDomain = math

        for domain_name, domain in self._domain_instances.items():
            # invariant from ``_DomainsContainer._from_environment``
            if domain_name != domain.name:
                msg = f'Domain name mismatch in {domain!r}: {domain_name!r} != {domain.name!r}'
                raise ValueError(msg)

    def _setup(self) -> None:
        for domain in self._domain_instances.values():
            domain.setup()

    def _process_doc(
        self, env: BuildEnvironment, docname: str, document: nodes.document
    ) -> None:
        for domain in self._domain_instances.values():
            domain.process_doc(env, docname, document)

    def _clear_doc(self, docname: str) -> None:
        for domain in self._domain_instances.values():
            domain.clear_doc(docname)

    def _merge_domain_data(
        self, docnames: Set[str], domain_data: dict[str, Any]
    ) -> None:
        for domain_name, domain in self._domain_instances.items():
            domain.merge_domaindata(docnames, domain_data[domain_name])

    def _check_consistency(self) -> None:
        for domain in self._domain_instances.values():
            domain.check_consistency()

    def __contains__(self, key: str) -> bool:
        return key in self._domain_instances

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _DomainsContainer):
            return NotImplemented
        return self._domain_instances == other._domain_instances

    def __hash__(self) -> int:
        return hash(sorted(self._domain_instances.items()))

    def __setattr__(self, key: str, value: object) -> None:
        if key in self._core_domains:
            msg = f'{self.__class__.__name__!r} object does not support assignment to {key!r}'
            raise TypeError(msg)
        super().__setattr__(key, value)

    def __delattr__(self, key: str) -> None:
        if key in self._core_domains:
            msg = f'{self.__class__.__name__!r} object does not support deletion of {key!r}'
            raise TypeError(msg)
        super().__delattr__(key)

    # Mapping interface: builtin domains

    @overload
    def __getitem__(self, key: Literal['c']) -> CDomain: ...

    @overload
    def __getitem__(self, key: Literal['cpp']) -> CPPDomain: ...

    @overload
    def __getitem__(self, key: Literal['changeset']) -> ChangeSetDomain: ...

    @overload
    def __getitem__(self, key: Literal['citation']) -> CitationDomain: ...

    @overload
    def __getitem__(self, key: Literal['index']) -> IndexDomain: ...

    @overload
    def __getitem__(self, key: Literal['js']) -> JavaScriptDomain: ...

    @overload
    def __getitem__(self, key: Literal['math']) -> MathDomain: ...

    @overload
    def __getitem__(self, key: Literal['py']) -> PythonDomain: ...

    @overload
    def __getitem__(self, key: Literal['rst']) -> ReSTDomain: ...

    @overload
    def __getitem__(self, key: Literal['std']) -> StandardDomain: ...

    # Mapping interface: first-party domains

    @overload
    def __getitem__(self, key: Literal['duration']) -> DurationDomain: ...

    @overload
    def __getitem__(self, key: Literal['todo']) -> TodoDomain: ...

    # Mapping interface: third-party domains

    @overload
    def __getitem__(self, key: str) -> Domain: ...

    def __getitem__(self, key: str) -> Domain:
        if domain := getattr(self, key, None):
            return domain
        return self._domain_instances[key]

    def __setitem__(self, key: str, value: Domain) -> NoReturn:
        msg = f'{self.__class__.__name__!r} object does not support item assignment'
        raise TypeError(msg)

    def __delitem__(self, key: str) -> NoReturn:
        msg = f'{self.__class__.__name__!r} object does not support item deletion'
        raise TypeError(msg)

    def __iter__(self) -> Iterator[str]:
        return iter(self._domain_instances.keys())

    def __len__(self) -> int:
        return len(self._domain_instances)

    def get(self, key: str, default: Domain | None = None) -> Domain | None:
        return self._domain_instances.get(key, default)

    def keys(self) -> Iterable[str]:
        return self._domain_instances.keys()

    def items(self) -> Iterable[tuple[str, Domain]]:
        return self._domain_instances.items()

    def values(self) -> Iterable[Domain]:
        return self._domain_instances.values()

    def sorted(self) -> Iterable[Domain]:
        for _domain_name, domain in sorted(self._domain_instances.items()):
            yield domain



from __future__ import annotations

import glob
import os
import os.path
from importlib.machinery import EXTENSION_SUFFIXES
from pathlib import Path
from typing import TYPE_CHECKING

from sphinx import package_dir
from sphinx.ext.apidoc._shared import LOGGER
from sphinx.locale import __
from sphinx.util.osutil import FileAvoidWrite
from sphinx.util.template import ReSTRenderer

if TYPE_CHECKING:
    import re
    from collections.abc import Iterable, Iterator, Sequence

    from sphinx.ext.apidoc._shared import ApidocOptions


# automodule options
if 'SPHINX_APIDOC_OPTIONS' in os.environ:
    OPTIONS = set(os.environ['SPHINX_APIDOC_OPTIONS'].split(','))
else:
    OPTIONS = {
        'members',
        'undoc-members',
        # 'inherited-members', # disabled because there's a bug in sphinx
        'show-inheritance',
    }

PY_SUFFIXES = ('.py', '.pyx', *EXTENSION_SUFFIXES)

template_dir = package_dir.joinpath('templates', 'apidoc')


def is_initpy(filename: str | Path) -> bool:
    """Check *filename* is __init__ file or not."""
    basename = Path(filename).name
    return any(
        basename == '__init__' + suffix
        for suffix in sorted(PY_SUFFIXES, key=len, reverse=True)
    )


def module_join(*modnames: str | None) -> str:
    """Join module names with dots."""
    return '.'.join(filter(None, modnames))


def is_package_dir(
    files: Iterable[Path | str] = (), *, dir_path: Path | None = None
) -> bool:
    """Check given *files* contains __init__ file."""
    if files != ():
        return any(map(is_initpy, files))
    if dir_path is not None:
        return any(map(is_initpy, dir_path.iterdir()))
    return False


def write_file(name: str, text: str, opts: ApidocOptions) -> Path:
    """Write the output file for module/package <name>."""
    fname = Path(opts.dest_dir, f'{name}.{opts.suffix}')
    if opts.dry_run:
        if not opts.quiet:
            LOGGER.info(__('Would create file %s.'), fname)
        return fname
    if not opts.force and fname.is_file():
        if not opts.quiet:
            LOGGER.info(__('File %s already exists, skipping.'), fname)
    else:
        if not opts.quiet:
            LOGGER.info(__('Creating file %s.'), fname)
        with FileAvoidWrite(fname) as f:
            f.write(text)
    return fname


def create_module_file(
    package: str | None,
    basename: str,
    opts: ApidocOptions,
    user_template_dir: str | os.PathLike[str] | None = None,
) -> Path:
    """Build the text of the file and write the file."""
    options = set(OPTIONS if not opts.automodule_options else opts.automodule_options)
    if opts.include_private:
        options.add('private-members')

    qualname = module_join(package, basename)
    context = {
        'show_headings': not opts.no_headings,
        'basename': basename,
        'qualname': qualname,
        'automodule_options': sorted(options),
    }
    template_path: Sequence[str | os.PathLike[str]]
    if user_template_dir is not None:
        template_path = [user_template_dir, template_dir]
    else:
        template_path = [template_dir]
    text = ReSTRenderer(template_path).render('module.rst.jinja', context)
    return write_file(qualname, text, opts)


def create_package_file(
    root: str,
    master_package: str | None,
    subroot: str,
    py_files: list[str],
    opts: ApidocOptions,
    subs: list[str],
    is_namespace: bool,
    excludes: Sequence[re.Pattern[str]] = (),
    user_template_dir: str | os.PathLike[str] | None = None,
) -> list[Path]:
    """Build the text of the file and write the file.

    Also create submodules if necessary.

    :returns: list of written files
    """
    # build a list of sub packages (directories containing an __init__ file)
    subpackages = [
        module_join(master_package, subroot, pkgname)
        for pkgname in subs
        if not is_skipped_package(Path(root, pkgname), opts, excludes)
    ]
    # build a list of sub modules
    submodules = [
        sub.split('.')[0]
        for sub in py_files
        if not is_skipped_module(Path(root, sub), opts, excludes) and not is_initpy(sub)
    ]
    submodules = sorted(set(submodules))
    submodules = [
        module_join(master_package, subroot, modname) for modname in submodules
    ]
    options = OPTIONS.copy()
    if opts.include_private:
        options.add('private-members')

    pkgname = module_join(master_package, subroot)
    context = {
        'pkgname': pkgname,
        'subpackages': subpackages,
        'submodules': submodules,
        'is_namespace': is_namespace,
        'modulefirst': opts.module_first,
        'separatemodules': opts.separate_modules,
        'automodule_options': sorted(options),
        'show_headings': not opts.no_headings,
        'maxdepth': opts.max_depth,
    }
    if user_template_dir is not None:
        template_path = [user_template_dir, template_dir]
    else:
        template_path = [template_dir]

    written: list[Path] = []

    text = ReSTRenderer(template_path).render('package.rst.jinja', context)
    written.append(write_file(pkgname, text, opts))

    if submodules and opts.separate_modules:
        written.extend([
            create_module_file(None, submodule, opts, user_template_dir)
            for submodule in submodules
        ])

    return written


def create_modules_toc_file(
    modules: list[str],
    opts: ApidocOptions,
    name: str = 'modules',
    user_template_dir: str | os.PathLike[str] | None = None,
) -> Path:
    """Create the module's index."""
    modules.sort()
    prev_module = ''
    for module in modules.copy():
        # look if the module is a subpackage and, if yes, ignore it
        if module.startswith(prev_module + '.'):
            modules.remove(module)
        else:
            prev_module = module

    context = {
        'header': opts.header,
        'maxdepth': opts.max_depth,
        'docnames': modules,
    }
    template_path: Sequence[str | os.PathLike[str]]
    if user_template_dir is not None:
        template_path = [user_template_dir, template_dir]
    else:
        template_path = [template_dir]
    text = ReSTRenderer(template_path).render('toc.rst.jinja', context)
    return write_file(name, text, opts)


def is_skipped_package(
    dirname: str | Path, opts: ApidocOptions, excludes: Sequence[re.Pattern[str]] = ()
) -> bool:
    """Check if we want to skip this module."""
    if not Path(dirname).is_dir():
        return False

    files = glob.glob(str(Path(dirname, '*.py')))  # NoQA: PTH207
    regular_package = any(f for f in files if is_initpy(f))
    if not regular_package and not opts.implicit_namespaces:
        # *dirname* is not both a regular package and an implicit namespace package
        return True

    # Check there is some showable module inside package
    return all(is_excluded(Path(dirname, f), excludes) for f in files)


def is_skipped_module(
    filename: str | Path, opts: ApidocOptions, _excludes: Sequence[re.Pattern[str]]
) -> bool:
    """Check if we want to skip this module."""
    filename = Path(filename)
    if not filename.exists():
        # skip if the file doesn't exist
        return True
    # skip if the module has a "private" name
    return filename.name.startswith('_') and not opts.include_private


def walk(
    root_path: str | Path,
    excludes: Sequence[re.Pattern[str]],
    opts: ApidocOptions,
) -> Iterator[tuple[str, list[str], list[str]]]:
    """Walk through the directory and list files and subdirectories up."""
    for root, subs, files in os.walk(root_path, followlinks=opts.follow_links):
        # document only Python module files (that aren't excluded)
        files = sorted(
            f
            for f in files
            if f.endswith(PY_SUFFIXES) and not is_excluded(Path(root, f), excludes)
        )

        # remove hidden ('.') and private ('_') directories, as well as
        # excluded dirs
        if opts.include_private:
            exclude_prefixes: tuple[str, ...] = ('.',)
        else:
            exclude_prefixes = ('.', '_')

        subs[:] = sorted(
            sub
            for sub in subs
            if not sub.startswith(exclude_prefixes)
            and not is_excluded(Path(root, sub), excludes)
        )

        yield root, subs, files


def has_child_module(
    root_path: str | Path, excludes: Sequence[re.Pattern[str]], opts: ApidocOptions
) -> bool:
    """Check the given directory contains child module/s (at least one)."""
    return any(files for _root, _subs, files in walk(root_path, excludes, opts))


def recurse_tree(
    root_path: str | os.PathLike[str],
    excludes: Sequence[re.Pattern[str]],
    opts: ApidocOptions,
    user_template_dir: str | os.PathLike[str] | None = None,
) -> tuple[list[Path], list[str]]:
    """Look for every file in the directory tree and create the corresponding
    ReST files.
    """
    # check if the base directory is a package and get its name
    root_path = Path(root_path)
    if is_package_dir(dir_path=root_path) or opts.implicit_namespaces:
        root_package = root_path.name
    else:
        # otherwise, the base is a directory with packages
        root_package = None

    toplevels = []
    written_files = []
    for root, subs, files in walk(root_path, excludes, opts):
        is_pkg = is_package_dir(files)
        is_namespace = not is_pkg and opts.implicit_namespaces
        if is_pkg:
            for f in files.copy():
                if is_initpy(f):
                    files.remove(f)
                    files.insert(0, f)
        elif root != str(root_path):
            # only accept non-package at toplevel unless using implicit namespaces
            if not opts.implicit_namespaces:
                subs.clear()
                continue

        if is_pkg or is_namespace:
            # we are in a package with something to document
            if subs or len(files) > 1 or not is_skipped_package(root, opts):
                subpackage = (
                    root.removeprefix(str(root_path))
                    .lstrip(os.path.sep)
                    .replace(os.path.sep, '.')
                )
                # if this is not a namespace or
                # a namespace and there is something there to document
                if not is_namespace or has_child_module(root, excludes, opts):
                    written_files.extend(
                        create_package_file(
                            root,
                            root_package,
                            subpackage,
                            files,
                            opts,
                            subs,
                            is_namespace,
                            excludes,
                            user_template_dir,
                        )
                    )
                    toplevels.append(module_join(root_package, subpackage))
        else:
            # if we are at the root level, we don't require it to be a package
            assert root == str(root_path)
            assert root_package is None
            for py_file in files:
                if not is_skipped_module(Path(root_path, py_file), opts, excludes):
                    module = py_file.split('.')[0]
                    written_files.append(
                        create_module_file(
                            root_package, module, opts, user_template_dir
                        )
                    )
                    toplevels.append(module)

    return written_files, toplevels


def is_excluded(root: str | Path, excludes: Sequence[re.Pattern[str]]) -> bool:
    """Check if the directory is in the exclude list.

    Note: by having trailing slashes, we avoid common prefix issues, like
          e.g. an exclude "foo" also accidentally excluding "foobar".
    """
    root_str = str(root)
    return any(exclude.match(root_str) for exclude in excludes)



"""Finnish search language."""

from __future__ import annotations

import snowballstemmer

from sphinx.search import SearchLanguage
from sphinx.search._stopwords.fi import FINNISH_STOPWORDS


class SearchFinnish(SearchLanguage):
    lang = 'fi'
    language_name = 'Finnish'
    js_stemmer_rawcode = 'finnish-stemmer.js'
    stopwords = FINNISH_STOPWORDS

    def __init__(self, options: dict[str, str]) -> None:
        super().__init__(options)
        self.stemmer = snowballstemmer.stemmer('finnish')

    def stem(self, word: str) -> str:
        return self.stemmer.stemWord(word.lower())



"""The citation domain."""

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from docutils import nodes

from sphinx.addnodes import pending_xref
from sphinx.domains import Domain
from sphinx.locale import __
from sphinx.transforms import SphinxTransform
from sphinx.util import logging
from sphinx.util.nodes import copy_source_info, make_refnode

if TYPE_CHECKING:
    from collections.abc import Set
    from typing import Any

    from docutils.nodes import Element

    from sphinx.application import Sphinx
    from sphinx.builders import Builder
    from sphinx.environment import BuildEnvironment
    from sphinx.util.typing import ExtensionMetadata


logger = logging.getLogger(__name__)


class CitationDomain(Domain):
    """Domain for citations."""

    name = 'citation'
    label = 'citation'

    dangling_warnings = {
        'ref': 'citation not found: %(target)s',
    }

    @property
    def citations(self) -> dict[str, tuple[str, str, int]]:
        return self.data.setdefault('citations', {})

    @property
    def citation_refs(self) -> dict[str, set[str]]:
        return self.data.setdefault('citation_refs', {})

    def clear_doc(self, docname: str) -> None:
        for key, (fn, _l, _lineno) in list(self.citations.items()):
            if fn == docname:
                del self.citations[key]
        for key, docnames in list(self.citation_refs.items()):
            if docnames == {docname}:
                del self.citation_refs[key]
            elif docname in docnames:
                docnames.remove(docname)

    def merge_domaindata(self, docnames: Set[str], otherdata: dict[str, Any]) -> None:
        # XXX duplicates?
        for key, data in otherdata['citations'].items():
            if data[0] in docnames:
                self.citations[key] = data
        for key, data in otherdata['citation_refs'].items():
            citation_refs = self.citation_refs.setdefault(key, set())
            for docname in data:
                if docname in docnames:
                    citation_refs.add(docname)

    def note_citation(self, node: nodes.citation) -> None:
        label = node[0].astext()
        if label in self.citations:
            path = self.env.doc2path(self.citations[label][0])
            logger.warning(
                __('duplicate citation %s, other instance in %s'),
                label,
                path,
                location=node,
                type='ref',
                subtype='citation',
            )
        self.citations[label] = (node['docname'], node['ids'][0], node.line)  # type: ignore[assignment]

    def note_citation_reference(self, node: pending_xref) -> None:
        docnames = self.citation_refs.setdefault(node['reftarget'], set())
        docnames.add(self.env.current_document.docname)

    def check_consistency(self) -> None:
        for name, (docname, _labelid, lineno) in self.citations.items():
            if name not in self.citation_refs:
                logger.warning(
                    __('Citation [%s] is not referenced.'),
                    name,
                    type='ref',
                    subtype='citation',
                    location=(docname, lineno),
                )

    def resolve_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        typ: str,
        target: str,
        node: pending_xref,
        contnode: Element,
    ) -> nodes.reference | None:
        docname, labelid, _lineno = self.citations.get(target, ('', '', 0))
        if not docname:
            return None

        return make_refnode(builder, fromdocname, docname, labelid, contnode)

    def resolve_any_xref(
        self,
        env: BuildEnvironment,
        fromdocname: str,
        builder: Builder,
        target: str,
        node: pending_xref,
        contnode: Element,
    ) -> list[tuple[str, nodes.reference]]:
        refnode = self.resolve_xref(
            env, fromdocname, builder, 'ref', target, node, contnode
        )
        if refnode is None:
            return []
        else:
            return [('ref', refnode)]


class CitationDefinitionTransform(SphinxTransform):
    """Mark citation definition labels as not smartquoted."""

    default_priority = 619

    def apply(self, **kwargs: Any) -> None:
        domain = self.env.domains.citation_domain
        for node in self.document.findall(nodes.citation):
            # register citation node to domain
            node['docname'] = self.env.current_document.docname
            domain.note_citation(node)

            # mark citation labels as not smartquoted
            label = cast('nodes.label', node[0])
            label['support_smartquotes'] = False


class CitationReferenceTransform(SphinxTransform):
    """Replace citation references by pending_xref nodes before the default
    docutils transform tries to resolve them.
    """

    default_priority = 619

    def apply(self, **kwargs: Any) -> None:
        domain = self.env.domains.citation_domain
        for node in self.document.findall(nodes.citation_reference):
            target = node.astext()
            ref = pending_xref(
                target,
                refdomain='citation',
                reftype='ref',
                reftarget=target,
                refwarn=True,
                support_smartquotes=False,
                ids=node['ids'],
                classes=node.get('classes', []),
            )
            ref += nodes.inline(target, '[%s]' % target)
            copy_source_info(node, ref)
            node.replace_self(ref)

            # register reference node to domain
            domain.note_citation_reference(ref)


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_domain(CitationDomain)
    app.add_transform(CitationDefinitionTransform)
    app.add_transform(CitationReferenceTransform)

    return {
        'version': 'builtin',
        'env_version': 1,
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""Update annotations info of living objects using type_comments."""

from __future__ import annotations

import ast
import contextlib
import sys
from inspect import Parameter, Signature, getsource
from types import ModuleType
from typing import TYPE_CHECKING, cast
from weakref import WeakSet

from sphinx.errors import PycodeError
from sphinx.ext.autodoc._shared import LOGGER
from sphinx.locale import __
from sphinx.pycode import ModuleAnalyzer
from sphinx.pycode.ast import unparse as ast_unparse
from sphinx.util import inspect
from sphinx.util.inspect import safe_getattr

if TYPE_CHECKING:
    from collections.abc import Sequence
    from typing import Any


_objects_with_type_comment_annotations: WeakSet[Any] = WeakSet()
"""Cache of objects with annotations updated from type comments."""


def _ensure_annotations_from_type_comments(obj: Any) -> None:
    """Ensures `obj.__annotations__` includes type comment information.

    Failures to assign to `__annotations__` are silently ignored.

    If `obj` is a class type, this also ensures that type comment
    information is incorporated into the `__annotations__` member of
    all parent classes, if possible.

    This mutates the `__annotations__` of existing imported objects,
    in order to allow the existing `typing.get_type_hints` method to
    take the modified annotations into account.

    Modifying existing imported objects is unfortunate but avoids the
    need to reimplement `typing.get_type_hints` in order to take into
    account type comment information.

    Note that this does not directly include type comment information
    from parent classes, but `typing.get_type_hints` takes that into
    account.
    """
    if obj in _objects_with_type_comment_annotations:
        return
    with contextlib.suppress(TypeError):
        _objects_with_type_comment_annotations.add(obj)

    if isinstance(obj, type):
        for cls in inspect.getmro(obj):
            modname = safe_getattr(cls, '__module__')
            mod = sys.modules.get(modname)
            if mod is not None:
                _ensure_annotations_from_type_comments(mod)

    elif isinstance(obj, ModuleType):
        _update_module_annotations_from_type_comments(obj)


def _update_module_annotations_from_type_comments(mod: ModuleType) -> None:
    """Adds type comment annotations for a single module.

    Both module-level and class-level annotations are added.
    """
    mod_annotations = dict(inspect.getannotations(mod))
    mod.__annotations__ = mod_annotations

    class_annotations: dict[str, dict[str, Any]] = {}

    try:
        analyzer = ModuleAnalyzer.for_module(mod.__name__)
        analyzer.analyze()
        anns = analyzer.annotations
        for (classname, attrname), annotation in anns.items():
            if not classname:
                annotations = mod_annotations
            else:
                cls_annotations = class_annotations.get(classname)
                if cls_annotations is None:
                    try:
                        cls = mod
                        for part in classname.split('.'):
                            cls = safe_getattr(cls, part)
                        annotations = dict(inspect.getannotations(cls))
                        # Ignore errors setting __annotations__
                        with contextlib.suppress(TypeError, AttributeError):
                            cls.__annotations__ = annotations
                    except AttributeError:
                        annotations = {}
                    class_annotations[classname] = annotations
                else:
                    annotations = cls_annotations
            annotations.setdefault(attrname, annotation)
    except PycodeError:
        pass


def _update_annotations_using_type_comments(obj: Any, bound_method: bool) -> None:
    """Update annotations info of *obj* using type_comments."""
    try:
        type_sig = get_type_comment(obj, bound_method)
        if type_sig:
            sig = inspect.signature(obj, bound_method)
            for param in sig.parameters.values():
                if param.name not in obj.__annotations__:
                    annotation = type_sig.parameters[param.name].annotation
                    if annotation is not Parameter.empty:
                        obj.__annotations__[param.name] = ast_unparse(annotation)

            if 'return' not in obj.__annotations__:
                obj.__annotations__['return'] = type_sig.return_annotation
    except KeyError as exc:
        LOGGER.warning(
            __('Failed to update signature for %r: parameter not found: %s'), obj, exc
        )
    except NotImplementedError as exc:  # failed to ast.unparse()
        LOGGER.warning(__('Failed to parse type_comment for %r: %s'), obj, exc)


def get_type_comment(obj: Any, bound_method: bool = False) -> Signature | None:
    """Get type_comment'ed FunctionDef object from living object.

    This tries to parse original code for living object and returns
    Signature for given *obj*.
    """
    try:
        source = getsource(obj)
        if source.startswith((' ', r'\t')):
            # subject is placed inside class or block.  To read its docstring,
            # this adds if-block before the declaration.
            module = ast.parse('if True:\n' + source, type_comments=True)
            subject = cast('ast.FunctionDef', module.body[0].body[0])  # type: ignore[attr-defined]
        else:
            module = ast.parse(source, type_comments=True)
            subject = cast('ast.FunctionDef', module.body[0])

        type_comment = getattr(subject, 'type_comment', None)
        if type_comment:
            function = ast.parse(type_comment, mode='func_type', type_comments=True)
            return signature_from_ast(subject, bound_method, function)
        else:
            return None
    except (OSError, TypeError):  # failed to load source code
        return None
    except SyntaxError:  # failed to parse type_comments
        return None


def signature_from_ast(
    node: ast.FunctionDef, bound_method: bool, type_comment: ast.FunctionDef
) -> Signature:
    """Return a Signature object for the given *node*.

    :param bound_method: Specify *node* is a bound method or not
    """
    params = []
    for arg in node.args.posonlyargs:
        param = Parameter(
            arg.arg,
            Parameter.POSITIONAL_ONLY,
            annotation=arg.type_comment,
        )
        params.append(param)

    for arg in node.args.args:
        param = Parameter(
            arg.arg,
            Parameter.POSITIONAL_OR_KEYWORD,
            annotation=arg.type_comment or Parameter.empty,
        )
        params.append(param)

    if node.args.vararg:
        param = Parameter(
            node.args.vararg.arg,
            Parameter.VAR_POSITIONAL,
            annotation=node.args.vararg.type_comment or Parameter.empty,
        )
        params.append(param)

    for arg in node.args.kwonlyargs:
        param = Parameter(
            arg.arg,
            Parameter.KEYWORD_ONLY,
            annotation=arg.type_comment or Parameter.empty,
        )
        params.append(param)

    if node.args.kwarg:
        param = Parameter(
            node.args.kwarg.arg,
            Parameter.VAR_KEYWORD,
            annotation=node.args.kwarg.type_comment or Parameter.empty,
        )
        params.append(param)

    # Remove first parameter when *obj* is bound_method
    if bound_method and params:
        params.pop(0)

    # merge type_comment into signature
    if not_suppressed(type_comment.argtypes):  # type: ignore[attr-defined]
        for i, param in enumerate(params):
            params[i] = param.replace(annotation=type_comment.argtypes[i])  # type: ignore[attr-defined]

    if node.returns:
        return Signature(params, return_annotation=node.returns)
    elif type_comment.returns:
        return Signature(params, return_annotation=ast_unparse(type_comment.returns))
    else:
        return Signature(params)


def not_suppressed(argtypes: Sequence[ast.expr] = ()) -> bool:
    """Check given *argtypes* is suppressed type_comment or not."""
    if len(argtypes) == 0:  # no argtypees
        return False
    if len(argtypes) == 1:
        arg = argtypes[0]
        if isinstance(arg, ast.Constant) and arg.value is ...:  # suppressed
            return False
    # not suppressed
    return True



"""Importer utilities for autodoc"""

from __future__ import annotations

import traceback
from importlib.machinery import EXTENSION_SUFFIXES
from typing import TYPE_CHECKING, NamedTuple

from sphinx.errors import PycodeError
from sphinx.ext.autodoc._dynamic._importer import (
    _find_type_stub_spec,  # NoQA: F401
    _reload_module,  # NoQA: F401
    _StubFileLoader,  # NoQA: F401
)
from sphinx.ext.autodoc._dynamic._importer import (
    _import_module as import_module,
)
from sphinx.ext.autodoc._dynamic._importer import _mangle_name as mangle
from sphinx.ext.autodoc._dynamic._member_finder import (
    _filter_enum_dict,
    _is_native_enum_api,  # NoQA: F401
    unmangle,
)
from sphinx.ext.autodoc._dynamic._mock import ismock, undecorate
from sphinx.ext.autodoc._shared import LOGGER
from sphinx.pycode import ModuleAnalyzer
from sphinx.util.inspect import (
    getannotations,
    getmro,
    getslots,
    isenumclass,
    safe_getattr,
)

if TYPE_CHECKING:
    from typing import Any, Protocol

    from sphinx.ext.autodoc._legacy_class_based._documenters import ObjectMember

    class _AttrGetter(Protocol):
        def __call__(self, obj: Any, name: str, default: Any = ..., /) -> Any: ...


_NATIVE_SUFFIXES: frozenset[str] = frozenset({'.pyx', *EXTENSION_SUFFIXES})


# Retained: legacy class-based
def import_object(
    modname: str,
    objpath: list[str],
    objtype: str = '',
    attrgetter: _AttrGetter = safe_getattr,
) -> Any:
    if objpath:
        LOGGER.debug('[autodoc] from %s import %s', modname, '.'.join(objpath))
    else:
        LOGGER.debug('[autodoc] import %s', modname)

    try:
        module = None
        exc_on_importing = None
        objpath = objpath.copy()
        while module is None:
            try:
                module = import_module(modname, try_reload=True)
                LOGGER.debug('[autodoc] import %s => %r', modname, module)
            except ImportError as exc:
                LOGGER.debug('[autodoc] import %s => failed', modname)
                exc_on_importing = exc
                if '.' in modname:
                    # retry with parent module
                    modname, name = modname.rsplit('.', 1)
                    objpath.insert(0, name)
                else:
                    raise

        obj = module
        parent = None
        object_name = None
        for attrname in objpath:
            parent = obj
            LOGGER.debug('[autodoc] getattr(_, %r)', attrname)
            mangled_name = mangle(obj, attrname)
            obj = attrgetter(obj, mangled_name)

            try:
                LOGGER.debug('[autodoc] => %r', obj)
            except TypeError:
                # fallback of failure on logging for broken object
                # See: https://github.com/sphinx-doc/sphinx/issues/9095
                LOGGER.debug('[autodoc] => %r', (obj,))

            object_name = attrname
        return [module, parent, object_name, obj]
    except (AttributeError, ImportError) as exc:
        if isinstance(exc, AttributeError) and exc_on_importing:
            # restore ImportError
            exc = exc_on_importing

        if objpath:
            errmsg = 'autodoc: failed to import %s %r from module %r' % (
                objtype,
                '.'.join(objpath),
                modname,
            )
        else:
            errmsg = f'autodoc: failed to import {objtype} {modname!r}'

        if isinstance(exc, ImportError):
            # import_module() raises ImportError having real exception obj and
            # traceback
            real_exc = exc.args[0]
            traceback_msg = traceback.format_exception(exc)
            if isinstance(real_exc, SystemExit):
                errmsg += (
                    '; the module executes module level statement '
                    'and it might call sys.exit().'
                )
            elif isinstance(real_exc, ImportError) and real_exc.args:
                errmsg += '; the following exception was raised:\n%s' % real_exc.args[0]
            else:
                errmsg += '; the following exception was raised:\n%s' % traceback_msg
        else:
            errmsg += (
                '; the following exception was raised:\n%s' % traceback.format_exc()
            )

        LOGGER.debug(errmsg)
        raise ImportError(errmsg) from exc


# Retained: legacy class-based
class Attribute(NamedTuple):
    name: str
    directly_defined: bool
    value: Any


# Retained: legacy class-based
def get_object_members(
    subject: Any,
    objpath: list[str],
    attrgetter: _AttrGetter,
    analyzer: ModuleAnalyzer | None = None,
) -> dict[str, Attribute]:
    """Get members and attributes of target object."""
    from sphinx.ext.autodoc._legacy_class_based._sentinels import INSTANCEATTR

    # the members directly defined in the class
    obj_dict = attrgetter(subject, '__dict__', {})

    members: dict[str, Attribute] = {}

    # enum members
    if isenumclass(subject):
        for name, defining_class, value in _filter_enum_dict(
            subject, attrgetter, obj_dict
        ):
            # the order of occurrence of *name* matches the subject's MRO,
            # allowing inherited attributes to be shadowed correctly
            if unmangled := unmangle(defining_class, name):
                members[unmangled] = Attribute(
                    name=unmangled,
                    directly_defined=defining_class is subject,
                    value=value,
                )

    # members in __slots__
    try:
        subject___slots__ = getslots(subject)
        if subject___slots__:
            from sphinx.ext.autodoc._legacy_class_based._sentinels import SLOTSATTR

            for name in subject___slots__:
                members[name] = Attribute(
                    name=name, directly_defined=True, value=SLOTSATTR
                )
    except (TypeError, ValueError):
        pass

    # other members
    for name in dir(subject):
        try:
            value = attrgetter(subject, name)
            directly_defined = name in obj_dict
            unmangled = unmangle(subject, name)
            if unmangled and unmangled not in members:
                members[unmangled] = Attribute(
                    name=unmangled, directly_defined=directly_defined, value=value
                )
        except AttributeError:
            continue

    # annotation only member (ex. attr: int)
    for cls in getmro(subject):
        for name in getannotations(cls):
            unmangled = unmangle(cls, name)
            if unmangled and unmangled not in members:
                members[unmangled] = Attribute(
                    name=unmangled, directly_defined=cls is subject, value=INSTANCEATTR
                )

    if analyzer:
        # append instance attributes (cf. self.attr1) if analyzer knows
        namespace = '.'.join(objpath)
        for ns, name in analyzer.find_attr_docs():
            if namespace == ns and name not in members:
                members[name] = Attribute(
                    name=name, directly_defined=True, value=INSTANCEATTR
                )

    return members


# Retained: legacy class-based
def get_class_members(
    subject: Any, objpath: Any, attrgetter: _AttrGetter, inherit_docstrings: bool = True
) -> dict[str, ObjectMember]:
    """Get members and attributes of target class."""
    from sphinx.ext.autodoc._legacy_class_based._documenters import ObjectMember
    from sphinx.ext.autodoc._legacy_class_based._sentinels import INSTANCEATTR

    # the members directly defined in the class
    obj_dict = attrgetter(subject, '__dict__', {})

    members: dict[str, ObjectMember] = {}

    # enum members
    if isenumclass(subject):
        for name, defining_class, value in _filter_enum_dict(
            subject, attrgetter, obj_dict
        ):
            # the order of occurrence of *name* matches the subject's MRO,
            # allowing inherited attributes to be shadowed correctly
            if unmangled := unmangle(defining_class, name):
                members[unmangled] = ObjectMember(
                    unmangled, value, class_=defining_class
                )

    # members in __slots__
    try:
        subject___slots__ = getslots(subject)
        if subject___slots__:
            from sphinx.ext.autodoc._legacy_class_based._sentinels import SLOTSATTR

            for name, docstring in subject___slots__.items():
                members[name] = ObjectMember(
                    name, SLOTSATTR, class_=subject, docstring=docstring
                )
    except (TypeError, ValueError):
        pass

    # other members
    for name in dir(subject):
        try:
            value = attrgetter(subject, name)
            if ismock(value):
                value = undecorate(value)

            unmangled = unmangle(subject, name)
            if unmangled and unmangled not in members:
                if name in obj_dict:
                    members[unmangled] = ObjectMember(unmangled, value, class_=subject)
                else:
                    members[unmangled] = ObjectMember(unmangled, value)
        except AttributeError:
            continue

    try:
        for cls in getmro(subject):
            try:
                modname = safe_getattr(cls, '__module__')
                qualname = safe_getattr(cls, '__qualname__')
                analyzer = ModuleAnalyzer.for_module(modname)
                analyzer.analyze()
            except AttributeError:
                qualname = None
                analyzer = None
            except PycodeError:
                analyzer = None

            # annotation only member (ex. attr: int)
            for name in getannotations(cls):
                unmangled = unmangle(cls, name)
                if unmangled and unmangled not in members:
                    if analyzer and (qualname, unmangled) in analyzer.attr_docs:
                        docstring = '\n'.join(analyzer.attr_docs[qualname, unmangled])
                    else:
                        docstring = None

                    members[unmangled] = ObjectMember(
                        unmangled, INSTANCEATTR, class_=cls, docstring=docstring
                    )

            # append or complete instance attributes (cf. self.attr1) if analyzer knows
            if analyzer:
                for (ns, name), docstring in analyzer.attr_docs.items():
                    if ns == qualname and name not in members:
                        # otherwise unknown instance attribute
                        members[name] = ObjectMember(
                            name,
                            INSTANCEATTR,
                            class_=cls,
                            docstring='\n'.join(docstring),
                        )
                    elif (
                        ns == qualname
                        and docstring
                        and isinstance(members[name], ObjectMember)
                        and not members[name].docstring
                    ):
                        if cls != subject and not inherit_docstrings:
                            # If we are in the MRO of the class and not the class itself,
                            # and we do not want to inherit docstrings, then skip setting
                            # the docstring below
                            continue
                        # attribute is already known, because dir(subject) enumerates it.
                        # But it has no docstring yet
                        members[name].docstring = '\n'.join(docstring)
    except AttributeError:
        pass

    return members



"""Helpers for AST (Abstract Syntax Tree)."""

from __future__ import annotations

import ast
from typing import TYPE_CHECKING, overload

if TYPE_CHECKING:
    from typing import NoReturn

OPERATORS: dict[type[ast.AST], str] = {
    ast.Add: '+',
    ast.And: 'and',
    ast.BitAnd: '&',
    ast.BitOr: '|',
    ast.BitXor: '^',
    ast.Div: '/',
    ast.FloorDiv: '//',
    ast.Invert: '~',
    ast.LShift: '<<',
    ast.MatMult: '@',
    ast.Mult: '*',
    ast.Mod: '%',
    ast.Not: 'not',
    ast.Pow: '**',
    ast.Or: 'or',
    ast.RShift: '>>',
    ast.Sub: '-',
    ast.UAdd: '+',
    ast.USub: '-',
}


@overload
def unparse(node: None, code: str = '') -> None: ...


@overload
def unparse(node: ast.AST, code: str = '') -> str: ...


def unparse(node: ast.AST | None, code: str = '') -> str | None:
    """Unparse an AST to string."""
    if node is None:
        return None
    elif isinstance(node, str):
        return node
    return _UnparseVisitor(code).visit(node)


# a greatly cut-down version of `ast._Unparser`
class _UnparseVisitor(ast.NodeVisitor):
    def __init__(self, code: str = '') -> None:
        self.code = code

    def _visit_op(self, node: ast.AST) -> str:
        return OPERATORS[node.__class__]

    for _op in OPERATORS:
        locals()[f'visit_{_op.__name__}'] = _visit_op

    def visit_arg(self, node: ast.arg) -> str:
        if node.annotation:
            return f'{node.arg}: {self.visit(node.annotation)}'
        else:
            return node.arg

    def _visit_arg_with_default(self, arg: ast.arg, default: ast.AST | None) -> str:
        """Unparse a single argument to a string."""
        name = self.visit(arg)
        if default:
            if arg.annotation:
                name += ' = %s' % self.visit(default)
            else:
                name += '=%s' % self.visit(default)
        return name

    def visit_arguments(self, node: ast.arguments) -> str:
        defaults: list[ast.expr | None] = list(node.defaults)
        positionals = len(node.args)
        posonlyargs = len(node.posonlyargs)
        positionals += posonlyargs
        for _ in range(len(defaults), positionals):
            defaults.insert(0, None)

        kw_defaults: list[ast.expr | None] = list(node.kw_defaults)
        for _ in range(len(kw_defaults), len(node.kwonlyargs)):
            kw_defaults.insert(0, None)

        args: list[str] = [
            self._visit_arg_with_default(arg, defaults[i])
            for i, arg in enumerate(node.posonlyargs)
        ]

        if node.posonlyargs:
            args.append('/')

        for i, arg in enumerate(node.args):
            args.append(self._visit_arg_with_default(arg, defaults[i + posonlyargs]))

        if node.vararg:
            args.append('*' + self.visit(node.vararg))

        if node.kwonlyargs and not node.vararg:
            args.append('*')
        for i, arg in enumerate(node.kwonlyargs):
            args.append(self._visit_arg_with_default(arg, kw_defaults[i]))

        if node.kwarg:
            args.append('**' + self.visit(node.kwarg))

        return ', '.join(args)

    def visit_Attribute(self, node: ast.Attribute) -> str:
        return f'{self.visit(node.value)}.{node.attr}'

    def visit_BinOp(self, node: ast.BinOp) -> str:
        # Special case ``**`` to not have surrounding spaces.
        if isinstance(node.op, ast.Pow):
            return ''.join(map(self.visit, (node.left, node.op, node.right)))
        return ' '.join(map(self.visit, (node.left, node.op, node.right)))

    def visit_BoolOp(self, node: ast.BoolOp) -> str:
        op = ' %s ' % self.visit(node.op)
        return op.join(self.visit(e) for e in node.values)

    def visit_Call(self, node: ast.Call) -> str:
        args = ', '.join(
            [self.visit(e) for e in node.args]
            + [f'{k.arg}={self.visit(k.value)}' for k in node.keywords],
        )
        return f'{self.visit(node.func)}({args})'

    def visit_Constant(self, node: ast.Constant) -> str:
        if node.value is Ellipsis:
            return '...'
        elif isinstance(node.value, (int, float, complex)):
            if self.code:
                return ast.get_source_segment(self.code, node) or repr(node.value)
            else:
                return repr(node.value)
        else:
            return repr(node.value)

    def visit_Dict(self, node: ast.Dict) -> str:
        keys = (self.visit(k) for k in node.keys if k is not None)
        values = (self.visit(v) for v in node.values)
        items = (k + ': ' + v for k, v in zip(keys, values, strict=True))
        return '{' + ', '.join(items) + '}'

    def visit_Lambda(self, node: ast.Lambda) -> str:
        return 'lambda %s: ...' % self.visit(node.args)

    def visit_List(self, node: ast.List) -> str:
        return '[' + ', '.join(self.visit(e) for e in node.elts) + ']'

    def visit_Name(self, node: ast.Name) -> str:
        return node.id

    def visit_Set(self, node: ast.Set) -> str:
        return '{' + ', '.join(self.visit(e) for e in node.elts) + '}'

    def visit_Slice(self, node: ast.Slice) -> str:
        if not node.lower and not node.upper and not node.step:
            # Empty slice with default values -> [:]
            return ':'

        start = self.visit(node.lower) if node.lower else ''
        stop = self.visit(node.upper) if node.upper else ''
        if not node.step:
            # Default step size -> [start:stop]
            return f'{start}:{stop}'

        step = self.visit(node.step) if node.step else ''
        return f'{start}:{stop}:{step}'

    def visit_Subscript(self, node: ast.Subscript) -> str:
        def is_simple_tuple(value: ast.expr) -> bool:
            return (
                isinstance(value, ast.Tuple)
                and bool(value.elts)
                and not any(isinstance(elt, ast.Starred) for elt in value.elts)
            )

        if is_simple_tuple(node.slice):
            elts = ', '.join(self.visit(e) for e in node.slice.elts)  # type: ignore[attr-defined]
            return f'{self.visit(node.value)}[{elts}]'
        return f'{self.visit(node.value)}[{self.visit(node.slice)}]'

    def visit_UnaryOp(self, node: ast.UnaryOp) -> str:
        # UnaryOp is one of {UAdd, USub, Invert, Not}, which refer to ``+x``,
        # ``-x``, ``~x``, and ``not x``. Only Not needs a space.
        if isinstance(node.op, ast.Not):
            return f'{self.visit(node.op)} {self.visit(node.operand)}'
        return f'{self.visit(node.op)}{self.visit(node.operand)}'

    def visit_Tuple(self, node: ast.Tuple) -> str:
        if len(node.elts) == 0:
            return '()'
        elif len(node.elts) == 1:
            return '(%s,)' % self.visit(node.elts[0])
        else:
            return '(' + ', '.join(self.visit(e) for e in node.elts) + ')'

    def visit_Starred(self, node: ast.Starred) -> str:
        return f'*{self.visit(node.value)}'

    def generic_visit(self, node: ast.AST) -> NoReturn:
        raise NotImplementedError('Unable to parse %s object' % type(node).__name__)



"""Templates utility functions for Sphinx."""

from __future__ import annotations

import os
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

from jinja2 import TemplateNotFound
from jinja2.loaders import BaseLoader
from jinja2.sandbox import SandboxedEnvironment

from sphinx import package_dir
from sphinx.jinja2glue import SphinxFileSystemLoader
from sphinx.locale import get_translator
from sphinx.util import rst, texescape

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from typing import Any

    from jinja2.environment import Environment

_TEMPLATES_PATH = package_dir / 'templates'
_LATEX_TEMPLATES_PATH = _TEMPLATES_PATH / 'latex'


class BaseRenderer:
    def __init__(self, loader: BaseLoader | None = None) -> None:
        self.env = SandboxedEnvironment(loader=loader, extensions=['jinja2.ext.i18n'])
        self.env.filters['repr'] = repr
        # ``install_gettext_translations`` is injected by the ``jinja2.ext.i18n`` extension
        self.env.install_gettext_translations(get_translator())  # type: ignore[attr-defined]

    def render(self, template_name: str, context: dict[str, Any]) -> str:
        return self.env.get_template(template_name).render(context)

    def render_string(self, source: str, context: dict[str, Any]) -> str:
        return self.env.from_string(source).render(context)


class FileRenderer(BaseRenderer):
    def __init__(self, search_path: Sequence[str | os.PathLike[str]]) -> None:
        if isinstance(search_path, (str, os.PathLike)):
            search_path = [search_path]
        else:
            # filter "None" paths
            search_path = list(filter(None, search_path))

        loader = SphinxFileSystemLoader(search_path)
        super().__init__(loader)

    @classmethod
    def render_from_file(
        cls: type[FileRenderer],
        filename: str | os.PathLike[str],
        context: dict[str, Any],
    ) -> str:
        filename = Path(filename)
        return cls((filename.parent,)).render(filename.name, context)


class SphinxRenderer(FileRenderer):
    def __init__(
        self, template_path: Sequence[str | os.PathLike[str]] | None = None
    ) -> None:
        if template_path is None:
            template_path = (_TEMPLATES_PATH,)
        super().__init__(template_path)

    @classmethod
    def render_from_file(
        cls: type[FileRenderer],
        filename: str | os.PathLike[str],
        context: dict[str, Any],
    ) -> str:
        return FileRenderer.render_from_file(filename, context)


class LaTeXRenderer(SphinxRenderer):
    def __init__(
        self,
        template_path: Sequence[str | os.PathLike[str]] | None = None,
        latex_engine: str | None = None,
    ) -> None:
        if template_path is None:
            template_path = (_LATEX_TEMPLATES_PATH,)
        super().__init__(template_path)

        # use texescape as escape filter
        escape = partial(texescape.escape, latex_engine=latex_engine)
        self.env.filters['e'] = escape
        self.env.filters['escape'] = escape
        self.env.filters['eabbr'] = texescape.escape_abbr

        # use JSP/eRuby like tagging instead because curly bracket; the default
        # tagging of jinja2 is not good for LaTeX sources.
        self.env.variable_start_string = '<%='
        self.env.variable_end_string = '%>'
        self.env.block_start_string = '<%'
        self.env.block_end_string = '%>'
        self.env.comment_start_string = '<#'
        self.env.comment_end_string = '#>'


class ReSTRenderer(SphinxRenderer):
    def __init__(
        self,
        template_path: Sequence[str | os.PathLike[str]] | None = None,
        language: str | None = None,
    ) -> None:
        super().__init__(template_path)

        # add language to environment
        self.env.extend(language=language)

        # use texescape as escape filter
        self.env.filters['e'] = rst.escape
        self.env.filters['escape'] = rst.escape
        self.env.filters['heading'] = rst.heading


class SphinxTemplateLoader(BaseLoader):
    """A loader supporting template inheritance"""

    def __init__(
        self,
        confdir: str | os.PathLike[str],
        templates_paths: Sequence[str | os.PathLike[str]],
        system_templates_paths: Sequence[str | os.PathLike[str]],
    ) -> None:
        self.loaders = []
        self.sysloaders = []

        conf_dir = Path(confdir)
        for templates_path in templates_paths:
            loader = SphinxFileSystemLoader(conf_dir / templates_path)
            self.loaders.append(loader)

        for templates_path in system_templates_paths:
            loader = SphinxFileSystemLoader(templates_path)
            self.loaders.append(loader)
            self.sysloaders.append(loader)

    def get_source(
        self,
        environment: Environment,
        template: str,
    ) -> tuple[str, str, Callable[[], bool]]:
        if template.startswith('!'):
            # search a template from ``system_templates_paths``
            loaders = self.sysloaders
            template = template[1:]
        else:
            loaders = self.loaders

        for loader in loaders:
            try:
                return loader.get_source(environment, template)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(template)
